# -*- coding: utf-8 -*-

# Copyright 2015 Spanish National Research Council
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from ooi.occi.core import attribute as attr
from ooi.occi.core import kind
from ooi.occi.core import link
from ooi.occi.core import mixin
from ooi.occi import helpers


class NetworkInterface(link.Link):
    attributes = attr.AttributeCollection(["occi.networkinterface.interface",
                                           "occi.networkinterface.mac",
                                           "occi.networkinterface.state"])
    kind = kind.Kind(helpers.build_scheme('infrastructure'),
                     'networkinterface', 'network link resource',
                     attributes, 'networklink/',
                     related=[link.Link.kind])

    def __init__(self, mixins, source, target, id=None, interface=None,
                 mac=None, state=None):

        super(NetworkInterface, self).__init__(None, mixins, source,
                                               target, id)

        self.attributes["occi.networkinterface.interface"] = (
            attr.InmutableAttribute("occi.networkinterface.interface",
                                    interface))
        self.attributes["occi.networkinterface.mac"] = attr.MutableAttribute(
            "occi.networkinterface.mac", mac)
        self.attributes["occi.networkinterface.state"] = (
            attr.InmutableAttribute("occi.networkinterface.state", state))

    @property
    def interface(self):
        return self.attributes["occi.networkinterface.interface"].value

    @property
    def mac(self):
        return self.attributes["occi.networkinterface.mac"].value

    @mac.setter
    def mac(self, value):
        self.attributes["occi.networkinterface.mac"].value = value

    @property
    def state(self):
        return self.attributes["occi.networkinterface.state"].value

ip_network_interface = mixin.Mixin(
    helpers.build_scheme("infrastructure/networkinterface"),
    "ipnetworkinterface", "IP Network interface Mixin",
    attributes=attr.AttributeCollection([
        "occi.networkinterface.address",
        "occi.networkinterface.gateway",
        "occi.networkinterface.allocation"]))



# -*- coding: utf-8 -*-

# Copyright 2015 Spanish National Research Council
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from ooi.occi.core import action
from ooi.occi.core import attribute as attr
from ooi.occi.core import kind
from ooi.occi.core import mixin
from ooi.occi.core import resource
from ooi.occi import helpers

up = action.Action(helpers.build_scheme('infrastructure/network/action'),
                   "up", "up network instance")

down = action.Action(helpers.build_scheme('infrastructure/network/action'),
                     "down", "down network instance")


class NetworkResource(resource.Resource):
    attributes = attr.AttributeCollection(["occi.network.vlan",
                                           "occi.network.label",
                                           "occi.network.state"])
    actions = (up, down)
    kind = kind.Kind(helpers.build_scheme('infrastructure'), 'network',
                     'network resource', attributes, 'network/',
                     actions=actions,
                     related=[resource.Resource.kind])

    def __init__(self, title, summary=None, id=None, vlan=None, label=None,
                 state=None, mixins=[]):
        super(NetworkResource, self).__init__(title, mixins, summary=summary,
                                              id=id)
        self.attributes["occi.network.vlan"] = attr.MutableAttribute(
            "occi.network.vlan", vlan)
        self.attributes["occi.network.label"] = attr.MutableAttribute(
            "occi.network.label", label)
        self.attributes["occi.network.state"] = attr.InmutableAttribute(
            "occi.network.state", state)

    @property
    def vlan(self):
        return self.attributes["occi.network.vlan"].value

    @vlan.setter
    def vlan(self, value):
        self.attributes["occi.network.vlan"].value = value

    @property
    def label(self):
        return self.attributes["occi.network.label"].value

    @label.setter
    def label(self, value):
        self.attributes["occi.network.label"].value = value

    @property
    def state(self):
        return self.attributes["occi.network.state"].value


ip_network = mixin.Mixin(helpers.build_scheme("infrastructure/network"),
                         "ipnetwork", "IP Networking Mixin",
                         attributes=attr.AttributeCollection([
                             "occi.network.address",
                             "occi.network.gateway",
                             "occi.network.allocation"]))



# -*- coding: utf-8 -*-

# Copyright 2015 LIP - INDIGO-DataCloud
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
import uuid

import mock

from ooi.api import helpers
from ooi.tests import base
from ooi.tests import fakes_network as fakes
from ooi import utils


class TestNovaNetOpenStackHelper(base.TestCase):
    def setUp(self):
        super(TestNovaNetOpenStackHelper, self).setUp()
        self.version = "version foo bar baz"
        self.helper = helpers.OpenStackHelper(None, self.version)

    @mock.patch.object(helpers.OpenStackHelper, "_get_req")
    @mock.patch.object(helpers.OpenStackHelper, "tenant_from_req")
    def test_list_networks_with_public(self, m_t, m_rq):
        id = uuid.uuid4().hex
        resp = fakes.create_fake_json_resp({"networks": [{"id": id}]}, 200)
        req_mock = mock.MagicMock()
        req_mock.get_response.return_value = resp
        resp_float = fakes.create_fake_json_resp(
            {"floating_ip_pools": [{"id": id}]}, 200
        )
        req_mock_float = mock.MagicMock()
        req_mock_float.get_response.return_value = resp_float
        m_rq.side_effect = [req_mock, req_mock_float]
        ret = self.helper.list_networks(None)
        self.assertEqual(2, ret.__len__())

    @mock.patch.object(helpers.OpenStackHelper, "_get_req")
    @mock.patch.object(helpers.OpenStackHelper, "tenant_from_req")
    def test_list_networks_with_no_public(self, m_t, m_rq):
        id = uuid.uuid4().hex
        resp = fakes.create_fake_json_resp({"networks": [{"id": id}]}, 200)
        req_mock = mock.MagicMock()
        req_mock.get_response.return_value = resp
        resp_float = fakes.create_fake_json_resp(
            {"floating_ip_pools": []}, 204
        )
        req_mock_float = mock.MagicMock()
        req_mock_float.get_response.return_value = resp_float
        m_rq.side_effect = [req_mock, req_mock_float]
        ret = self.helper.list_networks(None)
        self.assertEqual(1, ret.__len__())

    @mock.patch.object(helpers.OpenStackHelper, "_get_req")
    @mock.patch.object(helpers.OpenStackHelper, "tenant_from_req")
    def test_list_networks(self, m_t, m_rq):
        id = uuid.uuid4().hex
        tenant_id = uuid.uuid4().hex
        m_t.return_value = tenant_id
        resp = fakes.create_fake_json_resp({"networks": [{"id": id}]}, 200)
        resp_float = fakes.create_fake_json_resp(
            {"floating_ip_pools": [{"id": id}]}, 200
        )
        req_mock = mock.MagicMock()
        req_mock.get_response.return_value = resp
        req_mock_float = mock.MagicMock()
        req_mock_float.get_response.return_value = resp_float
        m_rq.side_effect = [req_mock, req_mock_float]
        ret = self.helper.list_networks(None)
        self.assertEqual(id, ret[0]['id'])
        self.assertEqual(
            {'method': 'GET',
             'path': '/%s/os-networks' % (tenant_id)},
            m_rq.call_args_list[0][1]
        )
        self.assertEqual(
            {'method': 'GET',
             'path': '/%s/os-floating-ip-pools' % (tenant_id)},
            m_rq.call_args_list[1][1]
        )

    def test_get_network_public(self):
        id = 'PUBLIC'
        ret = self.helper.get_network_details(None, id)
        self.assertEqual(id, ret["id"])

    @mock.patch.object(helpers.OpenStackHelper, "_get_req")
    @mock.patch.object(helpers.OpenStackHelper, "tenant_from_req")
    def test_get_network(self, m_t, m_rq):
        id = uuid.uuid4().hex
        address = uuid.uuid4().hex
        gateway = uuid.uuid4().hex
        label = "network11"
        tenant_id = uuid.uuid4().hex
        m_t.return_value = tenant_id
        resp = fakes.create_fake_json_resp(
            {"network": {"id": id, "label": label,
                         "cidr": address,
                         "gateway": gateway}}, 200
        )
        req_mock = mock.MagicMock()
        req_mock.get_response.return_value = resp
        m_rq.return_value = req_mock
        ret = self.helper.get_network_details(None, id)
        self.assertEqual(id, ret["id"])
        self.assertEqual(address, ret["address"])
        self.assertEqual(gateway, ret["gateway"])
        self.assertEqual(label, ret["name"])
        m_rq.assert_called_with(
            None, method="GET",
            path="/%s/os-networks/%s" % (tenant_id, id),
            )

    @mock.patch.object(helpers.OpenStackHelper, "_get_req")
    @mock.patch.object(helpers.OpenStackHelper, "tenant_from_req")
    def test_create_net(self, m_t, m_rq):
        tenant_id = uuid.uuid4().hex
        m_t.return_value = tenant_id
        name = "name_net"
        net_id = uuid.uuid4().hex
        cidr = "0.0.0.0"
        gateway = "0.0.0.1"
        parameters = {"label": name,
                      "cidr": cidr,
                      "gateway": gateway
                      }
        resp = fakes.create_fake_json_resp(
            {"network": {"id": net_id, "label": name,
                         "cidr": cidr,
                         "gateway": gateway}}, 200
        )
        req_mock = mock.MagicMock()
        req_mock.get_response.return_value = resp
        m_rq.return_value = req_mock
        ret = self.helper.create_network(None,
                                         name=name,
                                         cidr=cidr,
                                         gateway=gateway,
                                         )
        body = utils.make_body('network', parameters)
        m_rq.assert_called_with(
            None, method="POST",
            content_type='application/json',
            path="/%s/os-networks" % (tenant_id),
            body=json.dumps(body)
        )
        self.assertEqual(cidr, ret['address'])
        self.assertEqual(name, ret['name'])
        self.assertEqual(gateway, ret['gateway'])
        self.assertEqual(net_id, ret['id'])

    @mock.patch.object(helpers.OpenStackHelper, "_get_req")
    @mock.patch.object(helpers.OpenStackHelper, "tenant_from_req")
    def test_delete_net(self, m_t, m_rq):
        tenant_id = uuid.uuid4().hex
        m_t.return_value = tenant_id
        net_id = uuid.uuid4().hex
        req_mock = mock.MagicMock()
        req_mock.get_response.return_value = []
        m_rq.return_value = req_mock
        ret = self.helper.delete_network(None, net_id)
        self.assertEqual(ret, [])
        m_rq.assert_called_with(
            None, method="DELETE",
            path="/%s/os-networks/%s" % (tenant_id, net_id),
        )


# -*- coding: utf-8 -*-

# Copyright 2015 Spanish National Research Council
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import collections
import uuid

import mock

from ooi.api import helpers
from ooi.api import network_link as network_link_api
from ooi import exception
from ooi.occi.core import collection
from ooi.occi.infrastructure import compute
from ooi.occi.infrastructure import network
from ooi.occi.infrastructure import network_link
from ooi.openstack import network as os_network
from ooi.tests import base
from ooi.tests import fakes
from ooi.tests import fakes_network as fake_nets


class TestNetworkLinkController(base.TestController):
    def setUp(self):
        super(TestNetworkLinkController, self).setUp()
        self.controller = network_link_api.Controller(mock.MagicMock(), None)

    @mock.patch.object(helpers.OpenStackHelper, "list_compute_net_links")
    def test_index(self, mock_list):
        req = fake_nets.create_req_test(None, None)
        tenant = fake_nets.tenants['foo']
        os_link_list = fake_nets.network_links[tenant["id"]]
        links = []
        for os_link in os_link_list:
            l = fake_nets.fake_build_link(
                os_link['network_id'], os_link['instance_id'], os_link['ip'],
                mac=None, pool=os_link['pool'], state=os_link['status']
            )
            links.append(l)
        mock_list.return_value = links
        ret = self.controller.index(req)
        self.assertIsInstance(ret, collection.Collection)
        if tenant["name"] == "foo":
            for idx, ip in enumerate(os_link_list):
                if ip["instance_id"]:
                    self.assertIsInstance(ret.resources[idx],
                                          os_network.OSNetworkInterface)
        else:
            self.assertEqual([], ret.resources)
        mock_list.assert_called_with(req)

    @mock.patch.object(helpers.OpenStackHelper, "list_compute_net_links")
    def test_index_Empty(self, mock_list):
        req = fake_nets.create_req_test(None, None)
        links = []
        mock_list.return_value = links
        ret = self.controller.index(req)
        self.assertIsInstance(ret, collection.Collection)
        self.assertEqual(ret.resources.__len__(), 0)

    @mock.patch.object(helpers.OpenStackHelper,
                       "delete_port")
    @mock.patch.object(network_link_api.Controller,
                       "_get_interface_from_id")
    def test_delete_fixed(self, mock_get, mock_delete):
        net_id = uuid.uuid4().hex
        server_id = uuid.uuid4().hex
        server_addr = "192.168.253.1"
        link_id = "%s_%s_%s" % (
            server_id, net_id, server_addr)
        ip_ident = uuid.uuid4().hex

        class FakeNetworkLink(object):
            target = collections.namedtuple("Target", ["id"])(net_id)
            source = collections.namedtuple("Source", ["id"])(server_id)
            address = server_addr
            mac = 'foo'
            id = link_id
            ip_id = ip_ident

        mock_get.return_value = FakeNetworkLink()
        mock_delete.return_value = []
        ret = self.controller.delete(None, link_id)
        self.assertEqual([], ret)
        mock_get.assert_called_with(None, link_id)
        mock_delete.assert_called_with(None, server_id, ip_ident)

    @mock.patch.object(helpers.OpenStackHelper, "release_floating_ip")
    @mock.patch.object(helpers.OpenStackHelper, "remove_floating_ip")
    @mock.patch.object(network_link_api.Controller, "_get_interface_from_id")
    def test_delete_public(self, mock_get, mock_remove, mock_release):
        net_id = "PUBLIC"
        server_id = uuid.uuid4().hex
        server_addr = "192.168.253.1"
        link_id = "%s_%s_%s" % (server_id, net_id, server_addr)
        mac_id = uuid.uuid4().hex

        class FakeNetworkLink(object):
            target = collections.namedtuple("Target", ["id"])(net_id)
            source = collections.namedtuple("Source", ["id"])(server_id)
            address = server_addr
            mac = mac_id
            id = link_id
            ip_id = "foo"

        link = FakeNetworkLink()
        mock_get.return_value = link
        mock_release.return_value = None
        mock_remove.return_value = None

        ret = self.controller.delete(None, link.id)
        self.assertEqual([], ret)

        mock_get.assert_called_with(None, link.id)
        mock_remove.assert_called_with(None, link.source.id, link.address)
        mock_release.assert_called_with(None, link.ip_id)

    @mock.patch.object(helpers.OpenStackHelper, "get_compute_net_link")
    def test_show_complete(self, mock_get):
        os_link_list = fake_nets.network_links[fake_nets.tenants['foo']['id']]
        for os_link in os_link_list:
            mock_get.return_value = fake_nets.fake_build_link(
                os_link['network_id'], os_link['instance_id'], os_link['ip'],
                mac=None, pool=os_link['pool'], state=os_link['status']
            )
            link_id = '%s_%s_%s' % (
                os_link['instance_id'],
                os_link['network_id'],
                os_link['ip'])

            ret = self.controller.show(None, link_id)
            self.assertIsInstance(ret, os_network.OSNetworkInterface)
            self.assertEqual(os_link["ip"], ret.address)
            mock_get.assert_called_with(None, str(os_link['instance_id']),
                                        os_link['network_id'],
                                        os_link['ip'])

    @mock.patch.object(network_link_api.Controller, "_get_interface_from_id")
    def test_show(self, mock_get):
        mock_get.return_value = "foo"

        ret = self.controller.show(None, "bar")
        self.assertEqual("foo", ret)
        mock_get.assert_called_with(None, "bar")

    def test_get_interface_from_id_invalid(self):
        self.assertRaises(exception.LinkNotFound,
                          self.controller._get_interface_from_id,
                          None,
                          "foobarbaz")

    @mock.patch.object(helpers.OpenStackHelper, "get_server")
    def test_get_interface_from_id_invalid_no_matching_server(self, mock_get):
        mock_get.return_value = {"addresses": {"foo": [{"addr": "1.1.1.2"}]}}

        self.assertRaises(exception.LinkNotFound,
                          self.controller._get_interface_from_id,
                          None,
                          "%s_1.1.1.1" % uuid.uuid4().hex)

    @mock.patch.object(helpers.OpenStackHelper, "get_compute_net_link")
    def test_get_interface_from_id(self, mock_get_server):
        server_id = uuid.uuid4().hex
        net_id = uuid.uuid4().hex
        server_addr = "1.1.1.1"
        link_id = "%s_%s_%s" % (server_id, net_id, server_addr)
        mock_get_server.return_value = fake_nets.fake_build_link(
            net_id, server_id, server_addr
        )
        ret = self.controller.show(None, link_id)
        self.assertIsInstance(ret, os_network.OSNetworkInterface)
        mock_get_server.assert_called_with(None, server_id, net_id,
                                           server_addr)

    def test_get_network_link_resources_fixed(self):
        server_id = uuid.uuid4().hex
        net_id = uuid.uuid4().hex
        server_addr = "1.1.1.1"
        os_link = fake_nets.fake_build_link(
            net_id, server_id, server_addr
        )
        ret = network_link_api._get_network_link_resources([os_link])

        self.assertIsInstance(ret, list)
        self.assertIsInstance(ret[0], os_network.OSNetworkInterface)
        self.assertIsInstance(ret[0].source, compute.ComputeResource)
        self.assertIsInstance(ret[0].target, network.NetworkResource)
        self.assertEqual(ret[0].target.id, net_id)
        self.assertIsInstance(ret[0].ip_id, type(None))
        self.assertEqual(1, len(ret[0].mixins))
        self.assertIn(network_link.ip_network_interface, ret[0].mixins)

    def test_get_network_link_resources_public(self):
        server_id = uuid.uuid4().hex
        net_id = 'PUBLIC'
        server_addr = "1.1.1.1"
        os_link = fake_nets.fake_build_link(
            net_id, server_id, server_addr, pool='public'
        )
        ret = network_link_api._get_network_link_resources([os_link])

        self.assertIsInstance(ret, list)
        self.assertIsInstance(ret[0], os_network.OSNetworkInterface)
        self.assertIsInstance(ret[0].source, compute.ComputeResource)
        self.assertIsInstance(ret[0].target, network.NetworkResource)
        self.assertEqual(ret[0].target.id, net_id)
        self.assertIsInstance(ret[0].ip_id, type(None))
        self.assertEqual(2, len(ret[0].mixins))
        self.assertIn(network_link.ip_network_interface, ret[0].mixins)
        has_pool = False
        for m in ret[0].mixins:
            if isinstance(m, os_network.OSFloatingIPPool):
                has_pool = True
                break
        self.assertTrue(has_pool)

    def test_get_network_link_resources_empty(self):
        ret = network_link_api._get_network_link_resources(None)
        self.assertEqual(ret.__len__(), 0)

    @mock.patch.object(helpers.OpenStackHelper, "assign_floating_ip")
    @mock.patch("ooi.api.helpers.get_id_with_kind")
    def test_create_public(self, mock_get_id, mock_assign):
        server_id = uuid.uuid4().hex
        net_id = "PUBLIC"
        ip = '8.0.0.0'
        parameters = {
            "occi.core.target": net_id,
            "occi.core.source": server_id,
        }
        categories = {network_link.NetworkInterface.kind}
        req = fake_nets.create_req_test_occi(parameters, categories)
        mock_assign.return_value = fake_nets.fake_build_link(
            net_id, server_id, ip
        )
        mock_get_id.side_effect = [('', net_id), ('', server_id)]
        ret = self.controller.create(req)
        self.assertIsNotNone(ret)
        link = ret.resources.pop()
        self.assertIsInstance(link, os_network.OSNetworkInterface)
        self.assertIsInstance(link.source, compute.ComputeResource)
        self.assertIsInstance(link.target, network.NetworkResource)
        self.assertEqual(net_id, link.target.id)
        self.assertEqual(server_id, link.source.id)
        mock_assign.assert_called_with(mock.ANY, net_id, server_id, None)

    @mock.patch.object(helpers.OpenStackHelper, "create_port")
    @mock.patch("ooi.api.helpers.get_id_with_kind")
    def test_create_fixed(self, mock_get_id, mock_cre_port):
        server_id = uuid.uuid4().hex
        net_id = uuid.uuid4().hex
        mock_get_id.side_effect = [('', net_id), ('', server_id)]
        ip = '8.0.0.0'
        parameters = {
            "occi.core.target": net_id,
            "occi.core.source": server_id,
        }
        categories = {network_link.NetworkInterface.kind}
        req = fake_nets.create_req_test_occi(parameters, categories)
        mock_cre_port.return_value = fake_nets.fake_build_link(
            net_id, server_id, ip
        )
        ret = self.controller.create(req)
        self.assertIsNotNone(ret)
        link = ret.resources.pop()
        self.assertIsInstance(link, os_network.OSNetworkInterface)
        self.assertIsInstance(link.source, compute.ComputeResource)
        self.assertIsInstance(link.target, network.NetworkResource)
        self.assertEqual(net_id, link.target.id)
        self.assertEqual(server_id, link.source.id)
        mock_cre_port.assert_called_with(mock.ANY, net_id, server_id)

    @mock.patch("ooi.api.helpers.get_id_with_kind")
    @mock.patch("ooi.occi.validator.Validator")
    def test_create_invalid_ids(self, mock_validator, mock_get_id):
        tenant = fakes.tenants["foo"]
        req = self._build_req(tenant["id"])
        net_id = "foobarbaz"
        server_id = uuid.uuid4().hex
        obj = {
            "attributes": {
                "occi.core.target": net_id,
                "occi.core.source": server_id,
            }
        }

        req.get_parser = mock.MagicMock()
        mock_get_id.side_effect = exception.Invalid()
        req.get_parser.return_value.return_value.parse.return_value = obj
        mock_validator.validate.return_value = True
        self.assertRaises(exception.Invalid,
                          self.controller.create, req, None)

    @mock.patch.object(helpers.OpenStackHelper, "associate_floating_ip")
    @mock.patch.object(helpers.OpenStackHelper, "allocate_floating_ip")
    @mock.patch("ooi.api.helpers.get_id_with_kind")
    @mock.patch("ooi.occi.validator.Validator")
    def test_create_no_pool(self, mock_validator, mock_get_id, mock_allocate,
                            mock_associate):
        tenant = fakes.tenants["foo"]
        req = self._build_req(tenant["id"])
        net_id = "PUBLIC"
        server_id = uuid.uuid4().hex
        obj = {
            "attributes": {
                "occi.core.target": net_id,
                "occi.core.source": server_id,
            },
            "schemes": {
            }
        }
        ips = fakes.floating_ips[fakes.tenants["baz"]["id"]]

        for ip in ips:
            req.get_parser = mock.MagicMock()
            req.get_parser.return_value.return_value.parse.return_value = obj
            mock_validator.validate.return_value = True
            mock_allocate.return_value = ip
            mock_associate.return_value = None
            mock_get_id.side_effect = [('', net_id), ('', server_id)]

            ret = self.controller.create(req, None)
            link = ret.resources.pop()
            self.assertIsInstance(link, os_network.OSNetworkInterface)
            self.assertIsInstance(link.source, compute.ComputeResource)
            self.assertIsInstance(link.target, network.NetworkResource)
            self.assertEqual(net_id, link.target.id)
            self.assertEqual(server_id, link.source.id)

            mock_allocate.assert_called_with(mock.ANY, None)
            mock_associate.assert_called_with(mock.ANY, server_id, ip["ip"])

    @mock.patch.object(helpers.OpenStackHelper, "associate_floating_ip")
    @mock.patch.object(helpers.OpenStackHelper, "allocate_floating_ip")
    @mock.patch("ooi.api.helpers.get_id_with_kind")
    @mock.patch("ooi.occi.validator.Validator")
    def test_create_with_pool(self, mock_validator, mock_get_id, mock_allocate,
                              mock_associate):
        tenant = fakes.tenants["foo"]
        req = self._build_req(tenant["id"])
        net_id = "PUBLIC"
        server_id = uuid.uuid4().hex
        pool_name = "public"
        obj = {
            "attributes": {
                "occi.core.target": net_id,
                "occi.core.source": server_id,
            },
            "schemes": {
                os_network.OSFloatingIPPool.scheme: [pool_name],
            }
        }
        ips = fakes.floating_ips[fakes.tenants["baz"]["id"]]

        for ip in ips:
            req.get_parser = mock.MagicMock()
            req.get_parser.return_value.return_value.parse.return_value = obj
            mock_validator.validate.return_value = True
            mock_allocate.return_value = ip
            mock_associate.return_value = None
            mock_get_id.side_effect = [('', net_id), ('', server_id)]

            ret = self.controller.create(req, None)
            link = ret.resources.pop()
            self.assertIsInstance(link, os_network.OSNetworkInterface)
            self.assertIsInstance(link.source, compute.ComputeResource)
            self.assertIsInstance(link.target, network.NetworkResource)
            self.assertEqual(net_id, link.target.id)
            self.assertEqual(server_id, link.source.id)

            mock_allocate.assert_called_with(mock.ANY, pool_name)
            mock_associate.assert_called_with(mock.ANY, server_id, ip["ip"])



# -*- coding: utf-8 -*-

# Copyright 2015 Spanish National Research Council
# Copyright 2015 LIP - INDIGO-DataCloud
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
import re
import uuid

import webob.dec
import webob.exc

from ooi import utils
import ooi.wsgi


application_url = "https://foo.example.org:8774/ooiv1"

tenants = {
    "foo": {"id": uuid.uuid4().hex,
            "name": "foo"},
    "bar": {"id": uuid.uuid4().hex,
            "name": "bar"},
    "baz": {"id": uuid.uuid4().hex,
            "name": "baz"},
}

flavors = {
    1: {
        "id": 1,
        "name": "foo",
        "vcpus": 2,
        "ram": 256,
        "disk": 10,
    },
    2: {
        "id": 2,
        "name": "bar",
        "vcpus": 4,
        "ram": 2014,
        "disk": 20,
    }
}

images = {
    "foo": {
        "id": "foo",
        "name": "foo",
    },
    "bar": {
        "id": "bar",
        "name": "bar",
    }
}

volumes = {
    tenants["foo"]["id"]: [
        {
            "id": uuid.uuid4().hex,
            "displayName": "foo",
            "size": 2,
            "status": "available",
            "attachments": [],
        },
        {
            "id": uuid.uuid4().hex,
            "displayName": "bar",
            "size": 3,
            "status": "available",
            "attachments": [],
        },
        {
            "id": uuid.uuid4().hex,
            "displayName": "baz",
            "size": 5,
            "status": "available",
            "attachments": [],
        },
    ],
    tenants["bar"]["id"]: [],
    tenants["baz"]["id"]: [
        {
            "id": uuid.uuid4().hex,
            "displayName": "volume",
            "size": 5,
            "status": "in-use",
        },
    ],
}

pools = {
    tenants["foo"]["id"]: [
        {
            "id": "foo",
            "name": "foo",
        },
        {
            "id": "bar",
            "name": "bar",
        }
    ],
    tenants["bar"]["id"]: [],
    tenants["baz"]["id"]: [
        {
            "id": "public",
            "name": "public",
        },
    ],
}

linked_vm_id = uuid.uuid4().hex

allocated_ip = "192.168.253.23"

floating_ips = {
    tenants["foo"]["id"]: [],
    tenants["bar"]["id"]: [],
    tenants["baz"]["id"]: [
        {
            "fixed_ip": "10.0.0.2",
            "id": uuid.uuid4().hex,
            "instance_id": linked_vm_id,
            "ip": "192.168.253.1",
            "pool": pools[tenants["baz"]["id"]][0]["name"],
        },
        {
            "fixed_ip": None,
            "id": uuid.uuid4().hex,
            "instance_id": None,
            "ip": "192.168.253.2",
            "pool": pools[tenants["baz"]["id"]][0]["name"],
        },
    ],
}

networks = {
    tenants["foo"]["id"]: [],
    tenants["bar"]["id"]: [],
    tenants["baz"]["id"]: [
        {"id": uuid.uuid4().hex},
        {"id": uuid.uuid4().hex}
        ]
}

ports = {
    tenants["foo"]["id"]: [
        {
            "port_id": uuid.uuid4().hex,
            "fixed_ips":
                [{"ip_address": uuid.uuid4().hex}],
            "mac_addr": uuid.uuid4().hex,
            "port_state": "DOWN",
            "net_id": uuid.uuid4().hex,
            "server_id": linked_vm_id
        },
    ],
    tenants["bar"]["id"]: [],
    tenants["baz"]["id"]: [
        {
            "port_id": uuid.uuid4().hex,
            "fixed_ips": [
                {"ip_address": "192.168.253.1"}
            ],
            "mac_addr": uuid.uuid4().hex,
            "port_state": "ACTIVE",
            "net_id": uuid.uuid4().hex,
            "server_id": linked_vm_id
        },

    ],
}

servers = {
    tenants["foo"]["id"]: [
        {
            "id": uuid.uuid4().hex,
            "name": "foo",
            "flavor": {"id": flavors[1]["id"]},
            "image": {"id": images["foo"]["id"]},
            "status": "ACTIVE",
        },
        {
            "id": uuid.uuid4().hex,
            "name": "bar",
            "flavor": {"id": flavors[2]["id"]},
            "image": {"id": images["bar"]["id"]},
            "status": "SHUTOFF",
        },
        {
            "id": uuid.uuid4().hex,
            "name": "baz",
            "flavor": {"id": flavors[1]["id"]},
            "image": {"id": images["bar"]["id"]},
            "status": "ERROR",
        },
    ],
    tenants["bar"]["id"]: [],
    tenants["baz"]["id"]: [
        {
            "id": linked_vm_id,
            "name": "withvolume",
            "flavor": {"id": flavors[1]["id"]},
            "image": {"id": images["bar"]["id"]},
            "status": "ACTIVE",
            "os-extended-volumes:volumes_attached": [
                {"id": volumes[tenants["baz"]["id"]][0]["id"]}
            ],
            "addresses": {
                "private": [
                    {"addr": (
                        (ports[tenants["baz"]["id"]]
                         [0]["fixed_ips"][0]["ip_address"])
                    ),
                        "OS-EXT-IPS:type": "fixed",
                        "OS-EXT-IPS-MAC:mac_addr": (
                            ports[tenants["baz"]["id"]][0]["mac_addr"]
                        )
                    },
                    {"addr": floating_ips[tenants["baz"]["id"]][0]["ip"],
                     "OS-EXT-IPS:type": "floating",
                     "OS-EXT-IPS-MAC:mac_addr": "1234"},
                ]
            }
        }
    ],
}

# avoid circular definition of attachments
volumes[tenants["baz"]["id"]][0]["attachments"] = [{
    # how consistent can OpenStack be!
    # depending on using /servers/os-volume_attachments
    # or /os-volumes it will return different field names
    "server_id": servers[tenants["baz"]["id"]][0]["id"],
    "serverId": servers[tenants["baz"]["id"]][0]["id"],
    "attachment_id": uuid.uuid4().hex,
    "volumeId": volumes[tenants["baz"]["id"]][0]["id"],
    "volume_id": volumes[tenants["baz"]["id"]][0]["id"],
    "device": "/dev/vdb",
    "id": volumes[tenants["baz"]["id"]][0]["id"],
}]


def fake_query_results():
    cats = []
    # OCCI Core
    cats.append(
        'link; '
        'scheme="http://schemas.ogf.org/occi/core#"; '
        'class="kind"; title="link"; '
        'location="%s/link/"' % application_url)
    cats.append(
        'resource; '
        'scheme="http://schemas.ogf.org/occi/core#"; '
        'class="kind"; title="resource"; '
        'rel="http://schemas.ogf.org/occi/core#entity"; '
        'location="%s/resource/"' % application_url)
    cats.append(
        'entity; '
        'scheme="http://schemas.ogf.org/occi/core#"; '
        'class="kind"; title="entity"; '
        'location="%s/entity/"' % application_url)

    # OCCI Infrastructure Compute
    cats.append(
        'compute; '
        'scheme="http://schemas.ogf.org/occi/infrastructure#"; '
        'class="kind"; title="compute resource"; '
        'rel="http://schemas.ogf.org/occi/core#resource"; '
        'location="%s/compute/"' % application_url)
    cats.append(
        'start; '
        'scheme="http://schemas.ogf.org/occi/infrastructure/compute/action#"; '
        'class="action"; title="start compute instance"')
    cats.append(
        'stop; '
        'scheme="http://schemas.ogf.org/occi/infrastructure/compute/action#"; '
        'class="action"; title="stop compute instance"')
    cats.append(
        'restart; '
        'scheme="http://schemas.ogf.org/occi/infrastructure/compute/action#"; '
        'class="action"; title="restart compute instance"')
    cats.append(
        'suspend; '
        'scheme="http://schemas.ogf.org/occi/infrastructure/compute/action#"; '
        'class="action"; title="suspend compute instance"')

    # OCCI Templates
    cats.append(
        'os_tpl; '
        'scheme="http://schemas.ogf.org/occi/infrastructure#"; '
        'class="mixin"; title="OCCI OS Template"; '
        'location="%s/os_tpl/"' % application_url)
    cats.append(
        'resource_tpl; '
        'scheme="http://schemas.ogf.org/occi/infrastructure#"; '
        'class="mixin"; title="OCCI Resource Template"; '
        'location="%s/resource_tpl/"' % application_url)

    # OpenStack Images
    cats.append(
        'bar; '
        'scheme="http://schemas.openstack.org/template/os#"; '
        'class="mixin"; title="bar"; '
        'rel="http://schemas.ogf.org/occi/infrastructure#os_tpl"; '
        'location="%s/os_tpl/bar"' % application_url)
    cats.append(
        'foo; '
        'scheme="http://schemas.openstack.org/template/os#"; '
        'class="mixin"; title="foo"; '
        'rel="http://schemas.ogf.org/occi/infrastructure#os_tpl"; '
        'location="%s/os_tpl/foo"' % application_url)

    # OpenStack Flavors
    cats.append(
        '1; '
        'scheme="http://schemas.openstack.org/template/resource#"; '
        'class="mixin"; title="Flavor: foo"; '
        'rel="http://schemas.ogf.org/occi/infrastructure#resource_tpl"; '
        'location="%s/resource_tpl/1"' % application_url)
    cats.append(
        '2; '
        'scheme="http://schemas.openstack.org/template/resource#"; '
        'class="mixin"; title="Flavor: bar"; '
        'rel="http://schemas.ogf.org/occi/infrastructure#resource_tpl"; '
        'location="%s/resource_tpl/2"' % application_url)

    # OCCI Infrastructure Network
    cats.append(
        'network; '
        'scheme="http://schemas.ogf.org/occi/infrastructure#"; '
        'class="kind"; title="network resource"; '
        'rel="http://schemas.ogf.org/occi/core#resource"; '
        'location="%s/network/"' % application_url)
    cats.append(
        'ipnetwork; '
        'scheme="http://schemas.ogf.org/occi/infrastructure/network#"; '
        'class="mixin"; title="IP Networking Mixin"')
    cats.append(
        'up; '
        'scheme="http://schemas.ogf.org/occi/infrastructure/network/action#"; '
        'class="action"; title="up network instance"')
    cats.append(
        'down; '
        'scheme="http://schemas.ogf.org/occi/infrastructure/network/action#"; '
        'class="action"; title="down network instance"')
    cats.append(
        'networkinterface; '
        'scheme="http://schemas.ogf.org/occi/infrastructure#"; '
        'class="kind"; title="network link resource"; '
        'rel="http://schemas.ogf.org/occi/core#link"; '
        'location="%s/networklink/"' % application_url)
    cats.append(
        'ipnetworkinterface; '
        'scheme="http://schemas.ogf.org/occi/infrastructure/'
        'networkinterface#"; '
        'class="mixin"; title="IP Network interface Mixin"')

    # OCCI Infrastructure Storage
    cats.append(
        'storage; '
        'scheme="http://schemas.ogf.org/occi/infrastructure#"; '
        'class="kind"; title="storage resource"; '
        'rel="http://schemas.ogf.org/occi/core#resource"; '
        'location="%s/storage/"' % application_url)
    cats.append(
        'storagelink; '
        'scheme="http://schemas.ogf.org/occi/infrastructure#"; '
        'class="kind"; title="storage link resource"; '
        'rel="http://schemas.ogf.org/occi/core#link"; '
        'location="%s/storagelink/"' % application_url)
    cats.append(
        'offline; '
        'scheme="http://schemas.ogf.org/occi/infrastructure/storage/action#"; '
        'class="action"; title="offline storage instance"')
    cats.append(
        'online; '
        'scheme="http://schemas.ogf.org/occi/infrastructure/storage/action#"; '
        'class="action"; title="online storage instance"')
    cats.append(
        'backup; '
        'scheme="http://schemas.ogf.org/occi/infrastructure/storage/action#"; '
        'class="action"; title="backup storage instance"')
    cats.append(
        'resize; '
        'scheme="http://schemas.ogf.org/occi/infrastructure/storage/action#"; '
        'class="action"; title="resize storage instance"')
    cats.append(
        'snapshot; '
        'scheme="http://schemas.ogf.org/occi/infrastructure/storage/action#"; '
        'class="action"; title="snapshot storage instance"')

    # OpenStack contextualization
    cats.append(
        'user_data; '
        'scheme="http://schemas.openstack.org/compute/instance#"; '
        'class="mixin"; title="Contextualization extension - user_data"')
    cats.append(
        'public_key; '
        'scheme="http://schemas.openstack.org/instance/credentials#"; '
        'class="mixin"; title="Contextualization extension - public_key"')

    result = []
    for c in cats:
        result.append(("Category", c))
    return result


class FakeOpenStackFault(ooi.wsgi.Fault):
    _fault_names = {
        400: "badRequest",
        401: "unauthorized",
        403: "forbidden",
        404: "itemNotFound",
        405: "badMethod",
        406: "notAceptable",
        409: "conflictingRequest",
        413: "overLimit",
        415: "badMediaType",
        429: "overLimit",
        501: "notImplemented",
        503: "serviceUnavailable"}

    @webob.dec.wsgify()
    def __call__(self, req):
        code = self.wrapped_exc.status_int
        fault_name = self._fault_names.get(code)
        explanation = self.wrapped_exc.explanation
        fault_data = {
            fault_name: {
                'code': code,
                'message': explanation}}
        self.wrapped_exc.body = utils.utf8(json.dumps(fault_data))
        self.wrapped_exc.content_type = "application/json"
        return self.wrapped_exc


class FakeApp(object):
    """Poor man's fake application."""

    def __init__(self):
        self.routes = {}

        for tenant in tenants.values():
            path = "/%s" % tenant["id"]

            self._populate(path, "server", servers[tenant["id"]], actions=True)
            self._populate(path, "volume", volumes[tenant["id"]], "os-volumes")
            self._populate(path, "floating_ip_pool", pools[tenant["id"]],
                           "os-floating-ip-pools")
            self._populate(path, "floating_ip", floating_ips[tenant["id"]],
                           "os-floating-ips")
            self._populate_ports(path, servers[tenant["id"]],
                                 ports[tenant["id"]])
            # NOTE(aloga): dict_values un Py3 is not serializable in JSON
            self._populate(path, "image", list(images.values()))
            self._populate(path, "flavor", list(flavors.values()))
            self._populate_attached_volumes(path, servers[tenant["id"]],
                                            volumes[tenant["id"]])

    def _populate(self, path_base, obj_name, obj_list,
                  objs_path=None, actions=[]):
        objs_name = "%ss" % obj_name
        if objs_path:
            path = "%s/%s" % (path_base, objs_path)
        else:
            path = "%s/%s" % (path_base, objs_name)
        objs_details_path = "%s/detail" % path
        self.routes[path] = create_fake_json_resp({objs_name: obj_list})
        self.routes[objs_details_path] = create_fake_json_resp(
            {objs_name: obj_list})

        for o in obj_list:
            obj_path = "%s/%s" % (path, o["id"])
            self.routes[obj_path] = create_fake_json_resp({obj_name: o})

            if actions:
                action_path = "%s/action" % obj_path
                self.routes[action_path] = webob.Response(status=202)

    def _populate_attached_volumes(self, path, server_list, vol_list):
        for s in server_list:
            attachments = []
            if "os-extended-volumes:volumes_attached" in s:
                for attach in s["os-extended-volumes:volumes_attached"]:
                    for v in vol_list:
                        if attach["id"] == v["id"]:
                            attachments.append(v["attachments"][0])
            path_base = "%s/servers/%s/os-volume_attachments" % (path, s["id"])
            self.routes[path_base] = create_fake_json_resp(
                {"volumeAttachments": attachments}
            )
            for attach in attachments:
                obj_path = "%s/%s" % (path_base, attach["id"])
                self.routes[obj_path] = create_fake_json_resp(
                    {"volumeAttachment": attach})

    def _populate_ports(self, path, servers_list, ports_list):
        if servers_list:
            for p in ports_list:
                path_base = "%s/servers/%s/%s" % (
                    path,
                    servers_list[0]["id"],
                    "os-interface"
                )
                self.routes[path_base] = create_fake_json_resp(
                    {"interfaceAttachments": [p]})

    @webob.dec.wsgify()
    def __call__(self, req):
        if req.method == "GET":
            return self._do_get(req)
        elif req.method == "POST":
            return self._do_post(req)
        elif req.method == "DELETE":
            return self._do_delete(req)

    def _do_create_server(self, req):
        # TODO(enolfc): this should check the json is
        # semantically correct
        s = {"server": {"id": "foo",
                        "name": "foo",
                        "flavor": {"id": "1"},
                        "image": {"id": "2"},
                        "status": "ACTIVE"}}
        return create_fake_json_resp(s)

    def _do_create_volume(self, req):
        # TODO(enolfc): this should check the json is
        # semantically correct
        s = {"volume": {"id": "foo",
                        "displayName": "foo",
                        "size": 1,
                        "status": "on-line"}}
        return create_fake_json_resp(s)

    def _do_create_attachment(self, req):
        v = {"volumeAttachment": {"serverId": "foo",
                                  "volumeId": "bar",
                                  "device": "/dev/vdb"}}
        return create_fake_json_resp(v, 202)

    def _do_allocate_ip(self, req):
        tenant = req.path_info.split('/')[1]
        body = req.json_body.copy()
        pool = body.popitem()
        if pool[1]:
            for p in pools[tenant]:
                if p["name"] == pool[1]:
                    break
            else:
                exc = webob.exc.HTTPNotFound()
                return FakeOpenStackFault(exc)
        ip = {"floating_ip": {"ip": allocated_ip, "id": 1}}
        return create_fake_json_resp(ip, 202)

    def _do_create_port(self, req):
        req_content = req.path_info.split('/')
        tenant = req_content[1]
        server = req_content[3]
        body = req.json_body.copy()
        net = body["interfaceAttachment"]["net_id"]
        port = ports[tenant]
        p = {"interfaceAttachment": {
            "port_id": uuid.uuid4().hex,
            "fixed_ips":
                [{"ip_address":
                    port[0]["fixed_ips"]
                    [0]["ip_address"]
                  }],
            "mac_addr": port[0]["mac_addr"],
            "port_state": "DOWN",
            "net_id": net,
            "server_id": server
        }}
        return create_fake_json_resp(p, 200)

    def _do_post(self, req):
        if req.path_info.endswith("servers"):
            return self._do_create_server(req)
        if req.path_info.endswith("os-volumes"):
            return self._do_create_volume(req)
        elif req.path_info.endswith("action"):
            body = req.json_body.copy()
            action = body.popitem()
            if action[0] in ["os-start", "os-stop", "reboot",
                             "addFloatingIp", "removeFloatingIp"]:
                return self._get_from_routes(req)
        elif req.path_info.endswith("os-volume_attachments"):
            return self._do_create_attachment(req)
        elif req.path_info.endswith("os-floating-ips"):
            return self._do_allocate_ip(req)
        elif req.path_info.endswith("os-interface"):
            return self._do_create_port(req)
        raise Exception

    def _do_delete(self, req):
        self._do_get(req)
        tested_paths = {
            r"/[^/]+/servers/[^/]+/os-volume_attachments/[^/]+$": 202,
            r"/[^/]+/os-floating-ips/[^/]+$": 202,
            r"/[^/]+/servers/[^/]+$": 204,
            r"/[^/]+/os-volumes/[^/]+$": 204,
            r"/[^/]+/servers/[^/]+/os-interface/[^/]+$": 204,
        }
        for p, st in tested_paths.items():
            if re.match(p, req.path_info):
                return create_fake_json_resp({}, st)
        raise Exception

    def _do_get(self, req):
        return self._get_from_routes(req)

    def _get_from_routes(self, req):
        try:
            ret = self.routes[req.path_info]
        except KeyError:
            exc = webob.exc.HTTPNotFound()
            ret = FakeOpenStackFault(exc)
        return ret


def create_fake_json_resp(data, status=200):
    r = webob.Response()
    r.headers["Content-Type"] = "application/json"
    r.charset = "utf8"
    r.body = json.dumps(data).encode("utf8")
    r.status_code = status
    return r



# -*- coding: utf-8 -*-

# Copyright 2015 Spanish National Research Council
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc
import collections
import copy

import six


@six.add_metaclass(abc.ABCMeta)
class Attribute(object):
    def __init__(self, name, value):
        self._name = name
        self._value = value

    @property
    def name(self):
        return self._name

    @property
    def value(self):
        return self._value


class MutableAttribute(Attribute):
    @Attribute.value.setter
    def value(self, value):
        self._value = value


class InmutableAttribute(Attribute):
    pass


class AttributeCollection(object):
    def __init__(self, attributes=None):
        if attributes is not None:
            if isinstance(attributes, collections.Mapping):
                if not all([isinstance(a, Attribute)
                            for a in attributes.values()]):
                    raise TypeError('mapping keys must be of class Attribute')
                self.attributes = dict(attributes)
            elif isinstance(attributes, collections.Sequence):
                self.attributes = dict.fromkeys(attributes)
            else:
                raise TypeError('attributes must be a sequence or mapping.')
        else:
            self.attributes = {}

    def __getitem__(self, key):
        ret = self.attributes[self.__keytransform__(key)]
        if ret is None:
            raise AttributeError("attribute %s is not set" % key)
        return ret

    def __setitem__(self, key, value):
        self.attributes[self.__keytransform__(key)] = value

    def __delitem__(self, key):
        del self.attributes[self.__keytransform__(key)]

    def __iter__(self):
        return iter(self.attributes)

    def __len__(self):
        return len(self.attributes)

    def __keytransform__(self, key):
        return key

    def update(self, col):
        if not isinstance(col, AttributeCollection):
            raise TypeError('cannot update AttributeCollection with %s' %
                            type(col))
        return self.attributes.update(col.attributes)

    def copy(self):
        return copy.deepcopy(self)



# -*- coding: utf-8 -*-

# Copyright 2015 Spanish National Research Council
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

import mock

from ooi.api import helpers
from ooi.api import storage as storage_api
from ooi import exception
from ooi.occi.core import collection
from ooi.occi.infrastructure import storage
from ooi.tests import base
from ooi.tests import fakes


class TestStorageController(base.TestController):
    def setUp(self):
        super(TestStorageController, self).setUp()
        self.controller = storage_api.Controller(mock.MagicMock(), None)

    @mock.patch.object(helpers.OpenStackHelper, "get_volumes")
    def test_index(self, m_volumes):
        for tenant in fakes.tenants.values():
            vols = fakes.volumes[tenant["id"]]
            m_volumes.return_value = vols
            ret = self.controller.index(None)
            self.assertIsInstance(ret, collection.Collection)
            for idx, vol in enumerate(vols):
                self.assertIsInstance(ret.resources[idx],
                                      storage.StorageResource)
                self.assertEqual(vol["id"], ret.resources[idx].id)
                m_volumes.assert_called_with(None)

    @mock.patch.object(storage_api.Controller, "_delete")
    def test_delete(self, mock_delete):
        mock_delete.return_value = []
        ret = self.controller.delete(None, "foo")
        self.assertEqual([], ret)
        mock_delete.assert_called_with(None, ["foo"])

    @mock.patch.object(helpers.OpenStackHelper, "volume_delete")
    def test_delete_ids(self, mock_delete):
        vol_ids = [uuid.uuid4().hex, uuid.uuid4().hex]
        mock_delete.return_value = None
        ret = self.controller._delete(None, vol_ids)
        self.assertEqual([], ret)
        mock_delete.assert_has_calls([mock.call(None, s) for s in vol_ids])

    @mock.patch.object(helpers.OpenStackHelper, "get_volumes")
    @mock.patch.object(storage_api.Controller, "_delete")
    def test_delete_all(self, mock_delete, mock_get_volumes):
        vols = [{"id": uuid.uuid4().hex}, {"id": uuid.uuid4().hex}]
        mock_get_volumes.return_value = vols
        mock_delete.return_value = []
        ret = self.controller.delete_all(None)
        self.assertEqual([], ret)
        mock_delete.assert_called_with(None, [v["id"] for v in vols])

    @mock.patch.object(helpers.OpenStackHelper, "get_volume")
    def test_show(self, m_vol):
        for tenant in fakes.tenants.values():
            vols = fakes.volumes[tenant["id"]]
            for idx, vol in enumerate(vols):
                m_vol.return_value = vol
                ret = self.controller.show(None, vol["id"])[0]
                self.assertIsInstance(ret, storage.StorageResource)
                self.assertEqual(vol["id"], ret.id)
                self.assertEqual(vol["displayName"], ret.title)
                self.assertEqual(vol["size"], ret.size)
                m_vol.assert_called_with(None, vol["id"])

    @mock.patch.object(helpers.OpenStackHelper, "volume_create")
    @mock.patch("ooi.occi.validator.Validator")
    def test_volume_create(self, m_validator, m_create):
        tenant = fakes.tenants["foo"]
        req = self._build_req(tenant["id"])
        name = "foo volume"
        size = "10"
        obj = {
            "attributes": {
                "occi.core.title": name,
                "occi.storage.size": size,
            }
        }
        # NOTE(aloga): the mocked call is
        # "parser = req.get_parser()(req.headers, req.body)"
        req.get_parser = mock.MagicMock()
        # NOTE(aloga): MOG!
        req.get_parser.return_value.return_value.parse.return_value = obj
        m_validator.validate.return_value = True
        volume = {
            "id": uuid.uuid4().hex,
            "displayName": name,
            "size": size,
            "status": "ACTIVE",
        }
        m_create.return_value = volume
        ret = self.controller.create(req, None)
        self.assertIsInstance(ret, collection.Collection)
        m_create.assert_called_with(mock.ANY, name, size)

    @mock.patch("ooi.occi.validator.Validator")
    def test_volume_create_invalid(self, m_validator):
        tenant = fakes.tenants["foo"]
        req = self._build_req(tenant["id"])
        name = "foo volume"
        obj = {
            "attributes": {
                "occi.core.title": name,
            }
        }
        # NOTE(aloga): the mocked call is
        # "parser = req.get_parser()(req.headers, req.body)"
        req.get_parser = mock.MagicMock()
        # NOTE(aloga): MOG!
        req.get_parser.return_value.return_value.parse.return_value = obj
        m_validator.validate.return_value = True
        self.assertRaises(exception.Invalid,
                          self.controller.create,
                          req,
                          None)

    def test_actions(self):
        tenant = fakes.tenants["foo"]
        actions = ("online", "offline", "backup", "snapshot", "resize")
        for action in actions:
            req = self._build_req(tenant["id"], path="/foo?action=%s" % action)
            vol_id = uuid.uuid4().hex
            self.assertRaises(exception.NotImplemented,
                              self.controller.run_action,
                              req, vol_id, None)

    def test_invalid_action(self):
        tenant = fakes.tenants["foo"]
        actions = ("foo", "")
        for action in actions:
            req = self._build_req(tenant["id"], path="/foo?action=%s" % action)
            vol_id = uuid.uuid4().hex
            self.assertRaises(exception.InvalidAction,
                              self.controller.run_action,
                              req, vol_id, None)



# -*- coding: utf-8 -*-

# Copyright 2015 Spanish National Research Council
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from ooi.api import base
import ooi.api.helpers
from ooi import exception
from ooi.occi.core import collection
from ooi.occi.infrastructure import compute
from ooi.occi.infrastructure import storage
from ooi.occi.infrastructure import storage_link
from ooi.occi import validator as occi_validator


class Controller(base.Controller):
    def __init__(self, *args, **kwargs):
        super(Controller, self).__init__(*args, **kwargs)
        self.os_helper = ooi.api.helpers.OpenStackHelper(
            self.app,
            self.openstack_version
        )

    def index(self, req):
        volumes = self.os_helper.get_volumes(req)
        occi_link_resources = []
        for v in volumes:
            for attach in v["attachments"]:
                if attach:
                    c = compute.ComputeResource(title="Compute",
                                                id=attach["serverId"])
                    s = storage.StorageResource(title="Storage", id=v["id"])
                    l = storage_link.StorageLink(c, s,
                                                 deviceid=attach["device"])
                    occi_link_resources.append(l)

        return collection.Collection(resources=occi_link_resources)

    def _get_attachment_from_id(self, req, attachment_id):
        try:
            server_id, vol_id = attachment_id.split('_', 1)
        except ValueError:
            raise exception.LinkNotFound(link_id=attachment_id)

        vols = self.os_helper.get_server_volumes_link(req, server_id)
        for v in vols:
            if vol_id == v["volumeId"]:
                return v
        raise exception.LinkNotFound(link_id=attachment_id)

    def show(self, req, id):
        v = self._get_attachment_from_id(req, id)
        c = compute.ComputeResource(title="Compute", id=v["serverId"])
        s = storage.StorageResource(title="Storage", id=v["volumeId"])
        return [storage_link.StorageLink(c, s, deviceid=v["device"])]

    def create(self, req, body):
        parser = req.get_parser()(req.headers, req.body)
        scheme = {"category": storage_link.StorageLink.kind}
        obj = parser.parse()
        validator = occi_validator.Validator(obj)
        validator.validate(scheme)

        attrs = obj.get("attributes", {})
        _, vol_id = ooi.api.helpers.get_id_with_kind(
            req,
            attrs.get("occi.core.target"),
            storage.StorageResource.kind)
        _, server_id = ooi.api.helpers.get_id_with_kind(
            req,
            attrs.get("occi.core.source"),
            compute.ComputeResource.kind)
        device = attrs.get("occi.storagelink.deviceid", None)
        attachment = self.os_helper.create_server_volumes_link(req,
                                                               server_id,
                                                               vol_id,
                                                               dev=device)
        c = compute.ComputeResource(title="Compute", id=server_id)
        s = storage.StorageResource(title="Storage", id=vol_id)
        l = storage_link.StorageLink(c, s, deviceid=attachment["device"])
        return collection.Collection(resources=[l])

    def delete(self, req, id):
        v = self._get_attachment_from_id(req, id)
        self.os_helper.delete_server_volumes_link(req,
                                                  v["serverId"], v["volumeId"])
        return []



# -*- coding: utf-8 -*-

# Copyright 2015 Spanish National Research Council
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from ooi.occi.core import attribute
from ooi.occi.core import entity
from ooi.occi.core import kind
from ooi.occi.core import link
from ooi.occi import helpers


class Resource(entity.Entity):
    """OCCI Resource.

    The heart of the OCCI Core Model is the Resource type. Any resource exposed
    through OCCI is a Resource or a sub-type thereof. A resource can be e.g. a
    virtual machine, a job in a job submission system, a user, etc.

    The Resource type is complemented by the Link type which associates one
    Resource instance with another. The Link type contains a number of common
    attributes that Link sub-types inherit.
    """

    attributes = attribute.AttributeCollection(["occi.core.summary"])

    kind = kind.Kind(helpers.build_scheme('core'), 'resource',
                     'resource', attributes, 'resource/',
                     related=[entity.Entity.kind])

    def __init__(self, title, mixins, id=None, summary=None):
        super(Resource, self).__init__(title, mixins, id=id)
        self.attributes["occi.core.summary"] = attribute.MutableAttribute(
            "occi.core.summary", summary)
        self._links = []

    def __eq__(self, other):
        return all([self.attributes[i].value == other.attributes[i].value
                    for i in self.attributes])

    @property
    def links(self):
        return self._links

    def link(self, target, mixins=[]):
        l = link.Link("", mixins, self, target)
        self._links.append(l)

    def add_link(self, link):
        self._links.append(link)

    @property
    def summary(self):
        return self.attributes["occi.core.summary"].value

    @summary.setter
    def summary(self, value):
        self.attributes["occi.core.summary"].value = value



# -*- coding: utf-8 -*-

# Copyright 2015 Spanish National Research Council
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import collections

from ooi.occi.rendering import headers as header_rendering
from ooi.occi.rendering import text as text_rendering
from ooi.occi.rendering import urilist as urilist_rendering
from ooi.wsgi import utils


_MEDIA_TYPE_MAP = collections.OrderedDict([
    ('text/plain', 'text'),
    ('text/occi', 'header'),
    ('text/uri-list', 'uri-list')

])


class BaseSerializer(object):
    def __init__(self, env):
        self.env = env


class TextSerializer(BaseSerializer):
    def serialize(self, data):
        if not isinstance(data, list):
            data = [data]

        renderers = []
        for d in data:
            renderers.append(text_rendering.get_renderer(d))

        ret = "\n".join([r.render(env=self.env) for r in renderers])
        return None, utils.utf8(ret)


class HeaderSerializer(BaseSerializer):
    def serialize(self, data):
        if not isinstance(data, list):
            data = [data]

        renderers = []
        for d in data:
            renderers.append(header_rendering.get_renderer(d))

        # Header renderers will return a list, so we must flatten the results
        # before returning them
        headers = [i for r in renderers for i in r.render(env=self.env)]
        return headers, utils.utf8("")


class UriListSerializer(TextSerializer):
    # TODO(enolfc): this is mostly duplicated code.
    def serialize(self, data):
        if not isinstance(data, list):
            data = [data]

        renderers = []
        for d in data:
            renderers.append(urilist_rendering.get_renderer(d))

        ret = "\n".join([r.render(env=self.env) for r in renderers])
        return None, utils.utf8(ret)


_SERIALIZERS_MAP = {
    "text": TextSerializer,
    "header": HeaderSerializer,
    "uri-list": UriListSerializer,
}


def get_media_map():
    return _MEDIA_TYPE_MAP


def get_default_serializers():
    return _SERIALIZERS_MAP


def get_supported_content_types():
    return _MEDIA_TYPE_MAP.keys()



# -*- coding: utf-8 -*-

# Copyright 2015 LIP - INDIGO-DataCloud
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
import uuid

import webob.dec
import webob.exc

from ooi.api import network
from ooi.api import network_link
from ooi import wsgi


application_url = "https://foo.example.org:8774/ooiv1"

tenants = {
    "foo": {"id": uuid.uuid4().hex,
            "name": "foo"},
    "bar": {"id": uuid.uuid4().hex,
            "name": "bar"},
    "public": {"id": uuid.uuid4().hex,
               "name": "bar"},
}

subnets = [
    {
        "id": uuid.uuid4().hex,
        "name": "private-subnet",
        "cidr": "33.0.0.1/24",
        "ip_version": "IPv4",
        "gateway_ip": "33.0.0.1",
    },
    {
        "id": uuid.uuid4().hex,
        "name": "public-subnet",
        "cidr": "44.0.0.1/24",
        "ip_version": "IPv4",
        "gateway_ip": "44.0.0.1",
    },
]

networks = {
    tenants["bar"]["id"]: [],
    tenants["foo"]["id"]: [
        {
            "id": uuid.uuid4().hex,
            "name": "foo",
            "subnet_info": subnets[0],
            "status": "ACTIVE",
        },
        {
            "id": uuid.uuid4().hex,
            "name": "bar",
            "subnet_info": subnets[1],
            "status": "DOWN",
        },
    ]
}

networks_nova = {
    tenants["bar"]["id"]: [],
    tenants["foo"]["id"]: [
        {
            "id": uuid.uuid4().hex,
            "label": "foo",
            "gateway": "33.0.0.1",
            "cidr": "33.0.0.1/24",
            "status": "ACTIVE",
        },
        {
            "id": uuid.uuid4().hex,
            "label": "bar",
            "gateway": "44.0.0.1",
            "cidr": "44.0.0.1/24",
            "status": "DOWN",
        },
    ]
}

pools = {
    tenants["bar"]["id"]: [],
    tenants["foo"]["id"]: [
        {
            "id": uuid.uuid4().hex,
            "name": "foo",
        },
        {
            "id": uuid.uuid4().hex,
            "name": "bar",
        },
    ]
}

linked_vm_id = uuid.uuid4().hex
linked_net_id = uuid.uuid4().hex

allocated_ip = "192.168.253.23"

ports = {
    tenants["foo"]["id"]: [
        {"id": uuid.uuid4().hex,
         "device_id": uuid.uuid4().hex,
         "device_owner": uuid.uuid4().hex
         }
    ]
}

network_links = {
    tenants["bar"]["id"]: [],
    tenants["foo"]["id"]: [
        {
            "ip": "10.0.0.2",
            "id": uuid.uuid4().hex,
            "instance_id": linked_vm_id,
            "ip": "192.168.253.1",
            "network_id": linked_net_id,
            "pool": pools[tenants["foo"]["id"]][0]["name"],
            'status':'active'
        },
        {
            "ip": None,
            "id": uuid.uuid4().hex,
            "instance_id": None,
            "network_id": linked_net_id,
            "ip": "192.168.253.2",
            "pool": pools[tenants["foo"]["id"]][0]["name"],
            'status':'inactive'
        },
    ],
    tenants["public"]["id"]: [
        {
            "ip": "10.0.0.2",
            "id": uuid.uuid4().hex,
            "instance_id": linked_vm_id,
            "ip": "192.168.253.1",
            "network_id": 'PUBLIC',
            "pool": pools[tenants["foo"]["id"]][0]["name"],
            'status': 'active'
        },
    ],
}


def create_fake_json_resp(data, status=200):
    r = webob.Response()
    r.headers["Content-Type"] = "application/json"
    r.charset = "utf8"
    r.body = json.dumps(data).encode("utf8")
    r.status_code = status
    return r


def add_to_fake_json_resp(req, obj_name, data):

    body = json.loads(req.body)
    for element in data:
        body[obj_name].append(element)

    req.body = json.dumps(body).encode("utf8")

    return req


def create_req_test(params, schemes):
    headers = create_header(params, schemes)
    req = webob.Request.blank(path="")
    req.headers = headers
    return wsgi.Request(req.environ)


def create_header(params, schemes, project=None):
    headers = {}
    att = ""
    if params is not None:
        for k, v in params.items():
            att = "%s, %s=%s " % (att, k, v)
        headers["X_OCCI_Attribute"] = att
    if schemes is not None:
        sch = ""
        cat = ""
        for k, v in schemes.items():
            sch = "%s, %s:%s" % (sch, k, v)
            cat = "%s, %s%s" % (cat, k, v)
        headers["schemes"] = sch
        # headers['Category']= cat
    if project is not None:
        headers["X_PROJECT_ID"] = project
    return headers


def create_req_test_occi(params, category):
    headers = create_header_occi(params, category)
    req = webob.Request.blank(path="")
    req.headers = headers
    return wsgi.Request(req.environ)


def create_header_occi(params, category, project=None):
    headers = {}
    att = ""
    if params is not None:
        for k, v in params.items():
            att = "%s, %s=%s" % (att, k, v)
        headers["X_OCCI_Attribute"] = att
    if category is not None:
        cat = ""
        for c in category:
            cat = "%s%s; scheme=%s; class=%s, " % (
                cat,
                c.term, c.scheme, c.occi_class)
        headers['Category'] = cat[:-1]
    if project is not None:
        headers['X_PROJECT_ID'] = project
    return headers


def fake_build_link(net_id, compute_id, ip, mac=None,
                    pool=None, state='active'):
    link = {}
    link['mac'] = mac
    link['pool'] = pool
    link['network_id'] = net_id
    link['compute_id'] = compute_id
    link['ip'] = ip
    link['state'] = state
    return link


def fake_network_link_occi(os_list_net):
    list_links = []
    for l in os_list_net:
        if l['instance_id']:
            list_links.append(fake_build_link(l['network_id'],
                                              l['instance_id'], l['ip']))
    return network_link._get_network_link_resources(list_links)


def fake_build_net(name, ip_version=4, address='0.0.0.11', gateway='0.0.0.1',
                   id=33, state='active'):
    link = {}
    link['id'] = id
    link['name'] = name
    link['address'] = address
    link['gateway'] = gateway
    link['ip_version'] = ip_version
    link['state'] = state
    return link


def fake_network_occi(os_list_net):
    list_nets = []
    for n in os_list_net:
        list_nets.append(fake_build_net(n['name'], id=n['id']))
    return network.Controller._get_network_resources(list_nets)


def build_occi_network(network):
    name = network["name"]
    network_id = network["id"]
    subnet_info = network["subnet_info"]
    status = network["status"].upper()
    if status in ("ACTIVE",):
        status = "active"
    else:
        status = "inactive"

    app_url = application_url
    cats = []
    cats.append('network; '
                'scheme='
                '"http://schemas.ogf.org/occi/infrastructure#";'
                ' class="kind"; title="network resource";'
                ' rel='
                '"http://schemas.ogf.org/occi/core#resource";'
                ' location="%s/network/"' % app_url)
    cats.append('ipnetwork; '
                'scheme='
                '"http://schemas.ogf.org/occi/infrastructure/network#";'
                ' class="mixin"; title="IP Networking Mixin"')
    cats.append('osnetwork; '
                'scheme='
                '"http://schemas.openstack.org/infrastructure/network#";'
                ' class="mixin"; title="openstack network"')

    links = []
    links.append('<%s/network/%s?action=up>; '
                 'rel="http://schemas.ogf.org/occi/'
                 'infrastructure/network/action#up"' %
                 (application_url, network_id))
    links.append('<%s/network/%s?action=down>; '
                 'rel="http://schemas.ogf.org/occi/'
                 'infrastructure/network/action#down"' %
                 (application_url, network_id))

    attrs = [
        'occi.core.id="%s"' % network_id,
        'occi.core.title="%s"' % name,
        'occi.network.state="%s"' % status,
        'org.openstack.network.ip_version="%s"' % subnet_info["ip_version"],
        'occi.network.address="%s"' % subnet_info["cidr"],
        'occi.network.gateway="%s"' % subnet_info["gateway_ip"],
        ]
    result = []
    for c in cats:
        result.append(("Category", c))
    for a in attrs:
        result.append(("X-OCCI-Attribute", a))
    for l in links:
        result.append(("Link", l))
    return result


def build_occi_nova(network):
    name = network["label"]
    network_id = network["id"]
    gateway = network["gateway"]
    cidr = network["cidr"]
    status = "active"

    app_url = application_url
    cats = []
    cats.append('network; '
                'scheme='
                '"http://schemas.ogf.org/occi/infrastructure#";'
                ' class="kind"; title="network resource";'
                ' rel='
                '"http://schemas.ogf.org/occi/core#resource";'
                ' location="%s/network/"' % app_url)
    cats.append('ipnetwork; '
                'scheme='
                '"http://schemas.ogf.org/occi/infrastructure/network#";'
                ' class="mixin"; title="IP Networking Mixin"')
    cats.append('osnetwork; '
                'scheme='
                '"http://schemas.openstack.org/infrastructure/network#";'
                ' class="mixin"; title="openstack network"')

    links = []
    links.append('<%s/network/%s?action=up>; '
                 'rel="http://schemas.ogf.org/occi/'
                 'infrastructure/network/action#up"' %
                 (application_url, network_id))
    links.append('<%s/network/%s?action=down>; '
                 'rel="http://schemas.ogf.org/occi/'
                 'infrastructure/network/action#down"' %
                 (application_url, network_id))

    attrs = [
        'occi.core.id="%s"' % network_id,
        'occi.core.title="%s"' % name,
        'occi.network.state="%s"' % status,
        'occi.network.address="%s"' % cidr,
        'occi.network.gateway="%s"' % gateway,
        ]
    result = []
    for c in cats:
        result.append(("Category", c))
    for a in attrs:
        result.append(("X-OCCI-Attribute", a))
    for l in links:
        result.append(("Link", l))
    return result



# -*- coding: utf-8 -*-

# Copyright 2015 Spanish National Research Council
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from ooi.occi.core import action
from ooi.occi.core import attribute as attr
from ooi.occi.core import kind
from ooi.occi.core import resource
from ooi.occi import helpers

online = action.Action(helpers.build_scheme('infrastructure/storage/action'),
                       "online", "online storage instance")

offline = action.Action(helpers.build_scheme('infrastructure/storage/action'),
                        "offline", "offline storage instance")

backup = action.Action(helpers.build_scheme('infrastructure/storage/action'),
                       "backup", "backup storage instance")

snapshot = action.Action(helpers.build_scheme('infrastructure/storage/action'),
                         "snapshot", "snapshot storage instance")

resize = action.Action(helpers.build_scheme('infrastructure/storage/action'),
                       "resize", "resize storage instance")


class StorageResource(resource.Resource):
    attributes = attr.AttributeCollection(["occi.storage.size",
                                           "occi.storage.state"])
    actions = (online, offline, backup, snapshot, resize)
    kind = kind.Kind(helpers.build_scheme('infrastructure'), 'storage',
                     'storage resource', attributes, 'storage/',
                     actions=actions,
                     related=[resource.Resource.kind])

    def __init__(self, title, summary=None, id=None, size=None, state=None):
        mixins = []
        super(StorageResource, self).__init__(title, mixins, summary=summary,
                                              id=id)
        self.attributes["occi.storage.size"] = attr.MutableAttribute(
            "occi.storage.size", size)
        self.attributes["occi.storage.state"] = attr.InmutableAttribute(
            "occi.storage.state", state)

    @property
    def size(self):
        return self.attributes["occi.storage.size"].value

    @size.setter
    def size(self, value):
        self.attributes["occi.storage.size"].value = value

    @property
    def state(self):
        return self.attributes["occi.storage.state"].value



# -*- coding: utf-8 -*-

# Copyright 2015 Spanish National Research Council
# Copyright 2015 LIP - INDIGO-DataCloud
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import six
import six.moves.urllib.parse as urlparse


def utf8(value):
    """Try to turn a string into utf-8 if possible.

    Code is modified from the utf8 function in
    http://github.com/facebook/tornado/blob/master/tornado/escape.py

    """
    if isinstance(value, six.text_type):
        return value.encode('utf-8')
    assert isinstance(value, str)
    return value


def join_url(base, parts):
    """Join several parts into a url.

    :param base: the base url
    :parts: parts to join into the url
    """
    url = base
    if not isinstance(parts, (list, tuple)):
        parts = [parts]

    for p in parts:
        if p.startswith("/"):
            # We won't get an absolute url
            p = p[1:]
        url = urlparse.urljoin(url, p)
    return url


def make_body(resource, parameters):
    """Creates body request using the parameters.

    :param resource: kind of resource
    :param parameters: list of parameters
    """
    content = {}
    for key in parameters.keys():
        content[key] = parameters[key]
    if resource:
        body = {resource: content}
    else:
        body = content
    return body


def get_query_string(parameters):
    """Creates query string.

    :param parameters: list of parameters
    """
    if parameters is None:
        query_string = None
    else:
        query_string = urlparse.urlencode(parameters)
    return query_string


def translate_parameters(translation, parameters):
    """Translates parameter names to OS

    :param translation: list of names for translation
    :param parameters: list of parameters
    """
    if not parameters:
        return None
    out = {}
    for key in parameters.keys():
        if key in translation:
            if parameters[key]:
                out[translation[key]] = parameters[key]
    return out



# -*- coding: utf-8 -*-

# Copyright 2015 Spanish National Research Council
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import collections

import mock

from ooi import exception
from ooi.occi import validator
from ooi.tests import base


class TestValidator(base.TestCase):
    def test_category(self):
        mixins = collections.Counter()
        schemes = collections.defaultdict(list)
        pobj = {
            "kind": "compute",
            "category": "foo type",
            "mixins": mixins,
            "schemes": schemes,
        }
        cat = mock.MagicMock()
        cat.type_id = "foo type"

        scheme = {"category": cat}

        v = validator.Validator(pobj)
        self.assertTrue(v.validate(scheme))

    def test_category_missmatch(self):
        mixins = collections.Counter()
        schemes = collections.defaultdict(list)
        pobj = {
            "kind": "compute",
            "category": "foo bar baz",
            "mixins": mixins,
            "schemes": schemes,
        }
        cat = mock.MagicMock()
        cat.type_id = "foo type"

        scheme = {"category": cat}

        v = validator.Validator(pobj)
        self.assertRaises(exception.OCCISchemaMismatch, v.validate, scheme)

    def test_missing_category(self):
        mixins = collections.Counter()
        schemes = collections.defaultdict(list)
        pobj = {
            "kind": "compute",
            "mixins": mixins,
            "schemes": schemes,
        }
        cat = mock.MagicMock()
        cat.type_id = "foo type"

        scheme = {"category": cat}

        v = validator.Validator(pobj)
        self.assertRaises(exception.OCCIMissingType, v.validate, scheme)

    def test_mixins(self):
        mixins = collections.Counter()
        schemes = collections.defaultdict(list)
        mixins["http://example.com/scheme#foo"] += 1
        schemes["http://example.com/scheme#"].append("foo")
        pobj = {
            "kind": "compute",
            "category": "foo type",
            "mixins": mixins,
            "schemes": schemes,
        }

        mixin = mock.MagicMock()
        scheme = {"mixins": [mixin]}

        mixin.scheme = "http://example.com/scheme#"
        mixin.term = "foo"
        v = validator.Validator(pobj)
        self.assertTrue(v.validate(scheme))

    def test_optional_mixins(self):
        mixins = collections.Counter()
        schemes = collections.defaultdict(list)
        mixins["http://example.com/scheme#foo"] += 1
        schemes["http://example.com/scheme#"].append("foo")
        pobj = {
            "kind": "compute",
            "category": "foo type",
            "mixins": mixins,
            "schemes": schemes,
        }

        mixin = mock.MagicMock()
        mixin.scheme = "http://example.com/scheme#"
        mixin.term = "foo"
        scheme = {"optional_mixins": [mixin]}

        v = validator.Validator(pobj)
        self.assertTrue(v.validate(scheme))

    def test_mixin_schema_mismatch(self):
        mixins = collections.Counter()
        schemes = collections.defaultdict(list)
        mixins["http://example.com/scheme#foo"] += 1
        schemes["http://example.com/scheme#"].append("foo")
        pobj = {
            "kind": "compute",
            "category": "foo type",
            "mixins": mixins,
            "schemes": schemes,
        }

        mixin = mock.MagicMock()
        mixin.scheme = "http://example.com/foo#"
        mixin.term = "foo"
        scheme = {"mixins": [mixin]}

        v = validator.Validator(pobj)
        self.assertRaises(exception.OCCIMissingType, v.validate, scheme)

    def test_mixin_schema_mismatch_term(self):
        mixins = collections.Counter()
        schemes = collections.defaultdict(list)
        mixins["http://example.com/scheme#foo"] += 1
        schemes["http://example.com/scheme#"].append("foo")
        pobj = {
            "kind": "compute",
            "category": "foo type",
            "mixins": mixins,
            "schemes": schemes,
        }

        mixin = mock.MagicMock()
        mixin.scheme = "http://example.com/scheme#"
        mixin.term = "bar"
        scheme = {"mixins": [mixin]}

        v = validator.Validator(pobj)
        self.assertRaises(exception.OCCIMissingType, v.validate, scheme)

    def test_no_optional_mixins(self):
        mixins = collections.Counter()
        schemes = collections.defaultdict(list)
        pobj = {
            "kind": "compute",
            "category": "foo type",
            "mixins": mixins,
            "schemes": schemes,
        }

        mixin = mock.MagicMock()
        mixin.scheme = "http://example.com/scheme#"
        mixin.term = "foo"
        scheme = {"optional_mixins": [mixin]}

        v = validator.Validator(pobj)
        self.assertTrue(v.validate(scheme))

    def test_something_unexpected(self):
        mixins = collections.Counter()
        schemes = collections.defaultdict(list)
        mixins["http://example.com/scheme#foo"] += 1
        schemes["http://example.com/scheme#"].append("foo")
        pobj = {
            "kind": "compute",
            "category": "foo type",
            "mixins": mixins,
            "schemes": schemes,
        }

        v = validator.Validator(pobj)
        self.assertRaises(exception.OCCISchemaMismatch, v.validate, {})

    def test_optional_links(self):
        mixins = collections.Counter()
        schemes = collections.defaultdict(list)
        links = {"foo": {"rel": "http://example.com/scheme#foo"}}
        pobj = {
            "kind": "compute",
            "category": "foo type",
            "mixins": mixins,
            "schemes": schemes,
            "links": links
        }
        link = mock.MagicMock()
        link.type_id = "http://example.com/scheme#foo"
        scheme = {"optional_links": [link]}
        v = validator.Validator(pobj)
        self.assertTrue(v.validate(scheme))

    def test_optional_links_invalid(self):
        mixins = collections.Counter()
        schemes = collections.defaultdict(list)
        links = {"foo": {"rel": "http://example.com/scheme#foo"}}
        pobj = {
            "kind": "compute",
            "category": "foo type",
            "mixins": mixins,
            "schemes": schemes,
            "links": links
        }
        link = mock.MagicMock()
        link.type_id = "http://example.com/scheme#foo"
        scheme = {"optional_links": []}
        v = validator.Validator(pobj)
        self.assertRaises(exception.OCCISchemaMismatch, v.validate, scheme)



# -*- coding: utf-8 -*-

# Copyright 2015 Spanish National Research Council
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from ooi.occi.core import category


class Action(category.Category):
    """OCCI Action.

    An Action represents an invocable operation applicable to a resource
    instance.
    """

    def __init__(self, scheme, term, title, attributes=None, location=None):
        super(Action, self).__init__(scheme, term, title,
                                     attributes=attributes,
                                     location="?action=%s" % term)

    def _class_name(self):
        return "action"



# -*- coding: utf-8 -*-

# Copyright 2015 Spanish National Research Council
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

import six

from ooi.occi.core import attribute
from ooi.occi.core import kind
from ooi.occi.core import mixin
from ooi.occi import helpers
from ooi import utils


class EntityMeta(type):
    """Meta class for Entity classes.

    Following OCCI Core model, all the Entity subclasses will have its own
    attributes, as long as they parent's ones.

    For example the Entity class defines "occi.core.id" and "occi.core.title"
    attributes, and the resource Resource class (that is a subclass of Entity)
    defines "occi.core.summary" as attributes. Therefore, the Resource class
    and all the objects should have all three attributes.

    This metaclass does this, by updating the attributes to those of the base
    class.
    """
    def __new__(cls, name, bases, dct):
        for kls in bases:
            if "attributes" in vars(kls):
                dct["attributes"].update(kls.attributes)

        return super(EntityMeta, cls).__new__(cls, name, bases, dct)

    def __init__(self, *args):
        super(EntityMeta, self).__init__(*args)


@six.add_metaclass(EntityMeta)
class Entity(object):
    """OCCI Entity.

    Entity is an abstract type, which both Resource and Link inherit. Each
    sub-type of Entity is identified by a unique Kind instance
    """

    attributes = attribute.AttributeCollection(["occi.core.id",
                                                "occi.core.title"])

    kind = kind.Kind(helpers.build_scheme('core'), 'entity',
                     'entity', attributes, 'entity/')

    def __init__(self, title, mixins, id=None):
        helpers.check_type(mixins, mixin.Mixin)
        self.mixins = mixins

        # NOTE(aloga): we need a copy of the attributes, otherwise we will be
        # using the class ones instead of the object ones.
        self.attributes = self.attributes.copy()

        # damn, we're shading a builtin
        if id is None:
            id = uuid.uuid4().hex

        self.attributes["occi.core.id"] = attribute.InmutableAttribute(
            "occi.core.id", id)
        self.attributes["occi.core.title"] = attribute.MutableAttribute(
            "occi.core.title", title)

    @property
    def id(self):
        return self.attributes["occi.core.id"].value

    @property
    def title(self):
        return self.attributes["occi.core.title"].value

    @title.setter
    def title(self, value):
        self.attributes["occi.core.title"].value = value

    @property
    def location(self):
        return utils.join_url(self.kind.location, self.id)



# -*- coding: utf-8 -*-

# Copyright 2015 Spanish National Research Council
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from ooi.occi.core import action
from ooi.occi.core import attribute
from ooi.occi.core import category
from ooi.occi.core import entity
from ooi.occi.core import kind
from ooi.occi.core import link
from ooi.occi.core import mixin
from ooi.occi.core import resource
from ooi.tests import base


class TestAttributes(base.TestCase):
    def test_base(self):
        attr = attribute.Attribute("occi.foo.bar", "bar")
        self.assertEqual("bar", attr.value)
        self.assertEqual("occi.foo.bar", attr.name)

    def test_mutable(self):
        attr = attribute.MutableAttribute("occi.foo.bar", "bar")
        attr.value = "bazonk"
        self.assertEqual("bazonk", attr.value)

    def test_inmutable(self):
        attr = attribute.InmutableAttribute("occi.foo.bar", "bar")

        def set_val():
            attr.value = "bazonk"

        self.assertRaises(AttributeError, set_val)


class TestAttributeCollection(base.TestCase):
    def test_collection(self):
        col = attribute.AttributeCollection()
        self.assertEqual({}, col.attributes)

    def test_collection_raises_if_not_set(self):
        col = attribute.AttributeCollection(["foo"])
        self.assertRaises(AttributeError,
                          col.__getitem__,
                          "foo")

    def test_collection_from_seq(self):
        seq = ["foo", "bar"]
        col = attribute.AttributeCollection(seq)
        self.assertItemsEqual(seq, col.attributes.keys())

    def test_collection_from_map(self):
        mapping = {"foo": attribute.Attribute("occi.foo.bar", "crap")}
        col = attribute.AttributeCollection(mapping)
        self.assertEqual(mapping, col.attributes)

    def test_update(self):
        mapping1 = {"occi.foo.1": attribute.Attribute("occi.foo.1", "bar")}
        mapping2 = {"occi.foo.2": attribute.Attribute("occi.foo.2", "baz")}
        col1 = attribute.AttributeCollection(mapping1)
        col2 = attribute.AttributeCollection(mapping2)
        self.assertEqual(mapping1, col1.attributes)
        self.assertEqual(mapping2, col2.attributes)
        col1.update(col2)
        mapping1.update(mapping2)
        self.assertEqual(mapping1, col1.attributes)

    def test_update_invalid(self):
        mapping = {"occi.foo.1": attribute.Attribute("occi.foo.1", "bar")}
        col = attribute.AttributeCollection(mapping)
        self.assertRaises(TypeError,
                          col.update,
                          {"foo": "bar"})

    def test_collection_from_invalid_map(self):
        mapping = {"foo": "bar"}
        self.assertRaises(TypeError,
                          attribute.AttributeCollection,
                          mapping)

    def test_collection_from_invalid(self):
        mapping = 1
        self.assertRaises(TypeError,
                          attribute.AttributeCollection,
                          mapping)


class BaseTestCoreOCCICategory(base.TestCase):
    args = ("scheme", "term", "title")
    obj = category.Category

    def test_obj(self):
        cat = self.obj(*self.args)

        for i in self.args:
            self.assertEqual(i, getattr(cat, i))


class TestCoreOCCICategory(BaseTestCoreOCCICategory):
    pass


class TestCoreOCCIKind(BaseTestCoreOCCICategory):
    obj = kind.Kind

    def setUp(self):
        super(TestCoreOCCIKind, self).setUp()

    def test_obj(self):
        k = self.obj(*self.args)
        for i in (self.args):
            self.assertEqual(i, getattr(k, i))

    def test_actions(self):
        actions = [action.Action(None, None, None)]
        kind = self.obj(*self.args, actions=actions)

        for i in (self.args):
            self.assertEqual(i, getattr(kind, i))
        self.assertEqual(actions, kind.actions)

    def test_actions_empty(self):
        actions = []
        kind = self.obj(*self.args, actions=actions)

        for i in (self.args):
            self.assertEqual(i, getattr(kind, i))
        self.assertEqual(actions, kind.actions)

    def test_actions_invalid(self):
        actions = None
        self.assertRaises(TypeError,
                          self.obj,
                          *self.args,
                          actions=actions)

    def test_actions_invalid_list(self):
        actions = [None]
        self.assertRaises(TypeError,
                          self.obj,
                          *self.args,
                          actions=actions)

    def test_related(self):
        related = [self.obj(None, None, None)]
        kind = self.obj(*self.args, related=related)

        for i in (self.args):
            self.assertEqual(i, getattr(kind, i))
        self.assertEqual(related, kind.related)

    def test_related_empty(self):
        related = []
        kind = self.obj(*self.args, related=related)

        for i in (self.args):
            self.assertEqual(i, getattr(kind, i))
        self.assertEqual(related, kind.related)

    def test_related_invalid(self):
        related = None
        self.assertRaises(TypeError,
                          self.obj,
                          *self.args,
                          related=related)

    def test_related_invalid_list(self):
        related = [None]
        self.assertRaises(TypeError,
                          self.obj,
                          *self.args,
                          related=related)


class TestCoreOCCIMixin(TestCoreOCCIKind):
    obj = mixin.Mixin


class TestCoreOCCIAction(BaseTestCoreOCCICategory):
    obj = action.Action


class TestCoreOCCIEntity(base.TestCase):
    def test_resource_class(self):
        e = entity.Entity
        self.assertIn("occi.core.id", e.attributes)
        self.assertIn("occi.core.title", e.attributes)
        self.assertEqual([], e.kind.related)
        # TODO(aloga): We need to check that the attributes are actually set
        # after we get an object

    def test_entity(self):
        e = entity.Entity("bar", [])
        self.assertIsInstance(e.kind, kind.Kind)
        self.assertIn("occi.core.id", e.attributes)
        self.assertIn("occi.core.title", e.attributes)

        self.assertIs(e.id, e.attributes["occi.core.id"].value)

        def set_attr():
            e.attributes["occi.core.id"].value = "foo"
        self.assertRaises(AttributeError, set_attr)

        def set_attr_directly():
            e.id = "foo"
        self.assertRaises(AttributeError, set_attr_directly)

        e.title = "baz"
        self.assertEqual("baz", e.attributes["occi.core.title"].value)
        self.assertEqual("baz", e.title)
        self.assertIs(e.title, e.attributes["occi.core.title"].value)

        e.attributes["occi.core.title"].value = "bar"
        self.assertEqual("bar", e.attributes["occi.core.title"].value)
        self.assertEqual("bar", e.title)
        self.assertIs(e.title, e.attributes["occi.core.title"].value)

    def test_entities(self):
        e1 = entity.Entity("foo", [])
        e2 = entity.Entity("bar", [])
        self.assertNotEqual(e1.id, e2.id)


class TestCoreOCCIResource(base.TestCase):
    def test_resource_class(self):
        r = resource.Resource
        self.assertIn("occi.core.id", r.attributes)
        self.assertIn("occi.core.summary", r.attributes)
        self.assertIn("occi.core.title", r.attributes)
        self.assertIn(entity.Entity.kind, r.kind.related)
        # TODO(aloga): We need to check that the attributes are actually set
        # after we get an object

    def test_resource(self):
        id = uuid.uuid4().hex
        r = resource.Resource("bar", [], summary="baz", id=id)
        self.assertIsInstance(r.kind, kind.Kind)
        self.assertEqual("resource", r.kind.term)
        self.assertEqual("bar", r.title)
        self.assertEqual("baz", r.summary)
        self.assertEqual(id, r.id)
        self.assertIn(entity.Entity.kind, r.kind.related)
        r.summary = "bazonk"
        self.assertEqual("bazonk", r.summary)

    def test_resources_equal(self):
        id = uuid.uuid4().hex
        r = resource.Resource("bar", [], summary="baz", id=id)
        s = resource.Resource("bar", [], summary="baz", id=id)
        self.assertEqual(r, s)

    def test_valid_link(self):
        r1 = resource.Resource(None, [])
        r2 = resource.Resource(None, [])
        r1.link(r2)
        self.assertIsInstance(r1.links[0], link.Link)
        self.assertIs(r1, r1.links[0].source)
        self.assertIs(r2, r1.links[0].target)

    def test_mixins(self):
        m = mixin.Mixin(None, None, None)
        r = resource.Resource(None, [m], [])
        self.assertIsInstance(r.kind, kind.Kind)
        self.assertEqual([m], r.mixins)

    def test_invalid_mixins(self):
        self.assertRaises(TypeError,
                          resource.Resource,
                          None, ["foo"], None)


class TestCoreOCCILink(base.TestCase):
    def test_correct_link(self):
        resource_1 = resource.Resource(None, [], None)
        resource_2 = resource.Resource(None, [], None)
        resource_3 = resource.Resource(None, [], None)
        l = link.Link(None, [], resource_1, resource_2)
        self.assertIsInstance(l.kind, kind.Kind)
        self.assertEqual("link", l.kind.term)
        self.assertIs(resource_1, l.source)
        self.assertIs(resource_2, l.target)

        self.assertIs(resource_1, l.attributes["occi.core.source"].value)
        self.assertIs(resource_1, l.source)
        self.assertIs(resource_1, l.attributes["occi.core.source"].value)

        self.assertIs(resource_2, l.attributes["occi.core.target"].value)
        self.assertIs(resource_2, l.target)
        self.assertIs(resource_2, l.attributes["occi.core.target"].value)

        l.source = resource_3
        self.assertIs(resource_3, l.attributes["occi.core.source"].value)
        self.assertIs(resource_3, l.source)
        self.assertIs(resource_3, l.attributes["occi.core.source"].value)

        l.target = resource_1
        self.assertIs(resource_1, l.target)
        self.assertIs(resource_1, l.attributes["occi.core.target"].value)



# -*- coding: utf-8 -*-

# Copyright 2015 Spanish National Research Council
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from ooi.occi.core import attribute
from ooi.occi.infrastructure import templates
from ooi.openstack import helpers


class OpenStackOSTemplate(templates.OCCIOSTemplate):
    scheme = helpers.build_scheme("template/os")

    def __init__(self, uuid, name):
        location = "%s/%s" % (self._location, uuid)
        super(OpenStackOSTemplate, self).__init__(
            uuid,
            name,
            related=[templates.os_tpl],
            location=location)


class OpenStackResourceTemplate(templates.OCCIResourceTemplate):
    scheme = helpers.build_scheme("template/resource")

    def __init__(self, id, name, cores, memory, disk, ephemeral=0, swap=0):
        attrs = [
            attribute.InmutableAttribute("occi.compute.cores", cores),
            attribute.InmutableAttribute("occi.compute.memory", memory),
            attribute.InmutableAttribute("org.openstack.flavor.disk", disk),
            attribute.InmutableAttribute("org.openstack.flavor.ephemeral",
                                         ephemeral),
            attribute.InmutableAttribute("org.openstack.flavor.swap", swap),
            attribute.InmutableAttribute("org.openstack.flavor.name", name)
        ]

        attrs = attribute.AttributeCollection({a.name: a for a in attrs})

        location = "%s/%s" % (self._location, id)
        super(OpenStackResourceTemplate, self).__init__(
            id,
            "Flavor: %s" % name,
            related=[templates.resource_tpl],
            attributes=attrs,
            location=location)

    @property
    def cores(self):
        return self.attributes["occi.compute.cores"].value

    @property
    def memory(self):
        return self.attributes["occi.compute.memory"].value

    @property
    def disk(self):
        return self.attributes["org.openstack.flavor.disk"].value

    @property
    def ephemeral(self):
        return self.attributes["org.openstack.flavor.ephemeral"].value

    @property
    def swap(self):
        return self.attributes["org.openstack.flavor.swap"].value

    @property
    def name(self):
        return self.attributes["org.openstack.flavor.name"].value



# -*- coding: utf-8 -*-

# Copyright 2015 Spanish National Research Council
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import collections
import shlex

from ooi import exception


_MEDIA_TYPE_MAP = collections.OrderedDict([
    ('text/plain', 'text'),
    ('text/occi', 'header')
])


def _quoted_split(s, separator=',', quotes='"'):
    """Splits a string considering quotes.

    e.g. _quoted_split('a,"b,c",d') -> ['a', '"b,c"', 'd']
    """
    splits = []
    partial = []
    in_quote = None
    for c in s:
        if in_quote:
            if c == in_quote:
                in_quote = None
        else:
            if c in quotes:
                in_quote = c
        if not in_quote and c in separator:
            if partial:
                splits.append(''.join(partial))
            partial = []
        else:
            partial.append(c)
    if partial:
        splits.append(''.join(partial))
    return splits


def _split_unquote(s, separator="="):
    """Splits a string considering quotes and removing them in the result.

    e.g. _split_unquote('a="b=d"') -> ['a', 'b=d']
    """
    lex = shlex.shlex(s, posix=True)
    lex.commenters = ""
    lex.whitespace = separator
    lex.whitespace_split = True
    return list(lex)


class BaseParser(object):
    def __init__(self, headers, body):
        self.headers = headers
        self.body = body

    def parse(self):
        raise NotImplemented


class TextParser(BaseParser):
    def parse_categories(self, headers):
        kind = action = None
        mixins = collections.Counter()
        schemes = collections.defaultdict(list)
        try:
            categories = headers["Category"]
        except KeyError:
            raise exception.OCCIInvalidSchema("No categories")
        for ctg in _quoted_split(categories):
            ll = _quoted_split(ctg, "; ")
            d = {"term": ll[0]}  # assumes 1st element => term's value
            try:
                d.update(dict([_split_unquote(i) for i in ll[1:]]))
            except ValueError:
                raise exception.OCCIInvalidSchema("Unable to parse category")
            ctg_class = d.get("class", None)
            ctg_type = '%(scheme)s%(term)s' % d
            if ctg_class == "kind":
                if kind is not None:
                    raise exception.OCCIInvalidSchema("Duplicated Kind")
                kind = ctg_type
            elif ctg_class == "action":
                if action is not None:
                    raise exception.OCCIInvalidSchema("Duplicated action")
                action = ctg_type
            elif ctg_class == "mixin":
                mixins[ctg_type] += 1
            schemes[d["scheme"]].append(d["term"])
        if action and kind:
            raise exception.OCCIInvalidSchema("Action and kind together?")
        return {
            "category": kind or action,
            "mixins": mixins,
            "schemes": schemes,
        }

    def parse_attributes(self, headers):
        attrs = {}
        try:
            header_attrs = headers["X-OCCI-Attribute"]
            for attr in _quoted_split(header_attrs):
                l = _split_unquote(attr)
                attrs[l[0].strip()] = l[1]
        except KeyError:
            pass
        return attrs

    def parse_links(self, headers):
        links = {}
        try:
            header_links = headers["Link"]
        except KeyError:
            return links
        for link in _quoted_split(header_links):
            ll = _quoted_split(link, "; ")
            # remove the "<" and ">"
            if ll[0][1] != "<" and ll[0][-1] != ">":
                raise exception.OCCIInvalidSchema("Unable to parse link")
            link_dest = ll[0][1:-1]
            try:
                d = dict([_split_unquote(i) for i in ll[1:]])
            except ValueError:
                raise exception.OCCIInvalidSchema("Unable to parse link")
            links[link_dest] = d
        return links

    def _convert_to_headers(self):
        if not self.body:
            raise exception.OCCIInvalidSchema("No schema found")
        hdrs = collections.defaultdict(list)
        for l in self.body.splitlines():
            hdr, content = l.split(":", 1)
            hdrs[hdr].append(content)
        return {hdr: ','.join(hdrs[hdr]) for hdr in hdrs}

    def _parse(self, headers):
        obj = self.parse_categories(headers)
        obj['attributes'] = self.parse_attributes(headers)
        obj['links'] = self.parse_links(headers)
        return obj

    def parse(self):
        return self._parse(self._convert_to_headers())


class HeaderParser(TextParser):
    def parse(self):
        return self._parse(self.headers)


_PARSERS_MAP = {
    "text": TextParser,
    "header": HeaderParser,
}


def get_media_map():
    return _MEDIA_TYPE_MAP


def get_default_parsers():
    return _PARSERS_MAP


def get_supported_content_types():
    return _MEDIA_TYPE_MAP.keys()



# -*- coding: utf-8 -*-

# Copyright 2015 Spanish National Research Council
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

import webob.exc

import ooi.api.base
import ooi.api.helpers
from ooi import exception
from ooi.occi.core import collection
from ooi.occi.infrastructure import compute
from ooi.occi.infrastructure import network
from ooi.occi.infrastructure import storage
from ooi.occi.infrastructure import storage_link
from ooi.occi import validator as occi_validator
from ooi.openstack import contextualization
from ooi.openstack import helpers
from ooi.openstack import network as os_network
from ooi.openstack import templates


def _create_network_link(addr, comp, net_id):
    net = network.NetworkResource(title="network", id=net_id)
    return os_network.OSNetworkInterface(comp, net,
                                         addr["OS-EXT-IPS-MAC:mac_addr"],
                                         addr["addr"])


class Controller(ooi.api.base.Controller):
    def __init__(self, *args, **kwargs):
        super(Controller, self).__init__(*args, **kwargs)
        self.compute_actions = compute.ComputeResource.actions
        self.os_helper = ooi.api.helpers.OpenStackHelper(
            self.app,
            self.openstack_version
        )

    def _get_compute_resources(self, servers):
        occi_compute_resources = []
        if servers:
            for s in servers:
                s = compute.ComputeResource(title=s["name"], id=s["id"])
                occi_compute_resources.append(s)

        return occi_compute_resources

    def index(self, req):
        servers = self.os_helper.index(req)
        occi_compute_resources = self._get_compute_resources(servers)

        return collection.Collection(resources=occi_compute_resources)

    def run_action(self, req, id, body):
        action = req.GET.get("action", None)
        occi_actions = [a.term for a in compute.ComputeResource.actions]

        if action is None or action not in occi_actions:
            raise exception.InvalidAction(action=action)

        parser = req.get_parser()(req.headers, req.body)
        obj = parser.parse()

        server = self.os_helper.get_server(req, id)

        if action == "stop":
            scheme = {"category": compute.stop}
        elif action == "start":
            scheme = {"category": compute.start}
            if server["status"] == "SUSPENDED":
                action = "resume"
            elif server["status"] == "PAUSED":
                action = "unpause"
        elif action == "restart":
            scheme = {"category": compute.restart}
        elif action == "suspend":
            scheme = {"category": compute.suspend}
        else:
            raise exception.NotImplemented

        validator = occi_validator.Validator(obj)
        validator.validate(scheme)

        self.os_helper.run_action(req, action, id)
        return []

    def _build_block_mapping(self, req, obj):
        mappings = []
        for l in obj.get("links", {}).values():
            if l["rel"] == storage.StorageResource.kind.type_id:
                _, vol_id = ooi.api.helpers.get_id_with_kind(
                    req,
                    l.get("occi.core.target"),
                    storage.StorageResource.kind)
                mapping = {
                    "source_type": "volume",
                    "uuid": vol_id,
                    "delete_on_termination": False,
                }
                try:
                    mapping['device_name'] = l['occi.storagelink.deviceid']
                except KeyError:
                    pass
                mappings.append(mapping)
        # this needs to be there if we have a mapping
        if mappings:
            image = obj["schemes"][templates.OpenStackOSTemplate.scheme][0]
            mappings.insert(0, {
                "source_type": "image",
                "destination_type": "local",
                "boot_index": 0,
                "delete_on_termination": True,
                "uuid": image,
            })
        return mappings

    def _get_network_from_req(self, req, obj):
        networks = []
        for l in obj.get("links", {}).values():
            if l["rel"] == network.NetworkResource.kind.type_id:
                _, net_id = ooi.api.helpers.get_id_with_kind(
                    req,
                    l.get("occi.core.target"),
                    network.NetworkResource.kind)
                net = {'uuid': net_id}
                networks.append(net)
        return networks

    def create(self, req, body):
        parser = req.get_parser()(req.headers, req.body)
        scheme = {
            "category": compute.ComputeResource.kind,
            "mixins": [
                templates.OpenStackOSTemplate,
                templates.OpenStackResourceTemplate,
            ],
            "optional_mixins": [
                contextualization.user_data,
                contextualization.public_key,
            ],
            "optional_links": [
                storage.StorageResource.kind,
                network.NetworkResource.kind,
            ]

        }
        obj = parser.parse()
        validator = occi_validator.Validator(obj)
        validator.validate(scheme)

        attrs = obj.get("attributes", {})
        name = attrs.get("occi.core.title", "OCCI_VM")
        image = obj["schemes"][templates.OpenStackOSTemplate.scheme][0]
        flavor = obj["schemes"][templates.OpenStackResourceTemplate.scheme][0]
        user_data, key_name, key_data = None, None, None
        create_key, create_key_tmp = False, False
        if contextualization.user_data.scheme in obj["schemes"]:
            user_data = attrs.get("org.openstack.compute.user_data")
        if contextualization.public_key.scheme in obj["schemes"]:
            key_name = attrs.get("org.openstack.credentials.publickey.name")
            key_data = attrs.get("org.openstack.credentials.publickey.data")

            if key_name and key_data:
                create_key = True
            elif not key_name and key_data:
                # NOTE(orviz) To be occi-os compliant, not
                # raise exception.MissingKeypairName
                key_name = uuid.uuid4().hex
                create_key = True
                create_key_tmp = True

            if create_key:
                # add keypair: if key_name already exists, a 409 HTTP code
                # will be returned by OpenStack
                self.os_helper.keypair_create(req, key_name,
                                              public_key=key_data)

        block_device_mapping_v2 = self._build_block_mapping(req, obj)
        networks = self._get_network_from_req(req, obj)
        server = self.os_helper.create_server(
            req,
            name,
            image,
            flavor,
            user_data=user_data,
            key_name=key_name,
            block_device_mapping_v2=block_device_mapping_v2,
            networks=networks
        )
        # The returned JSON does not contain the server name
        server["name"] = name
        occi_compute_resources = self._get_compute_resources([server])

        if create_key_tmp:
            self.os_helper.keypair_delete(req, key_name)

        return collection.Collection(resources=occi_compute_resources)

    def show(self, req, id):
        # get info from server
        s = self.os_helper.get_server(req, id)

        # get info from flavor
        flavor = self.os_helper.get_flavor(req, s["flavor"]["id"])
        res_tpl = templates.OpenStackResourceTemplate(flavor["id"],
                                                      flavor["name"],
                                                      flavor["vcpus"],
                                                      flavor["ram"],
                                                      flavor["disk"])

        # get info from image
        img_id = s["image"]["id"]
        try:
            image = self.os_helper.get_image(req, img_id)
        except webob.exc.HTTPNotFound:
            image = {
                "id": img_id,
                "name": "None (Image with ID '%s' not found)" % img_id,
            }

        os_tpl = templates.OpenStackOSTemplate(image["id"],
                                               image["name"])

        # build the compute object
        comp = compute.ComputeResource(title=s["name"], id=s["id"],
                                       cores=flavor["vcpus"],
                                       hostname=s["name"],
                                       memory=flavor["ram"],
                                       state=helpers.vm_state(s["status"]),
                                       mixins=[os_tpl, res_tpl])

        # storage links
        vols = self.os_helper.get_server_volumes_link(req, s["id"])
        for v in vols:
            st = storage.StorageResource(title="storage", id=v["volumeId"])
            comp.add_link(storage_link.StorageLink(comp, st,
                                                   deviceid=v["device"]))

        # network links
        addresses = s.get("addresses", {})
        if addresses:
            for addr_set in addresses.values():
                for addr in addr_set:
                    # TODO(jorgesece): add pool information
                    if addr["OS-EXT-IPS:type"] == "floating":
                        net_id = helpers.PUBLIC_NETWORK
                    else:
                        try:
                            net_id = self.os_helper.get_network_id(
                                req, addr['OS-EXT-IPS-MAC:mac_addr'], id
                            )
                        except webob.exc.HTTPNotFound:
                            net_id = "FIXED"
                    comp.add_link(_create_network_link(addr, comp, net_id))

        return [comp]

    def _get_server_floating_ips(self, req, server_id):
        s = self.os_helper.get_server(req, server_id)
        addresses = s.get("addresses", {})
        floating_ips = []
        if addresses:
            for addr_set in addresses.values():
                for addr in addr_set:
                    if addr["OS-EXT-IPS:type"] == "floating":
                        floating_ips.append(addr["addr"])
        return floating_ips

    def _release_floating_ips(self, req, server_id):
        server_ips = self._get_server_floating_ips(req, server_id)
        if server_ips:
            floating_ips = self.os_helper.get_floating_ips(req)
            for server_ip in server_ips:
                for ip in floating_ips:
                    if server_ip == ip["ip"]:
                        self.os_helper.remove_floating_ip(req, server_id,
                                                          ip["ip"])
                        self.os_helper.release_floating_ip(req, ip["id"])

    def _delete(self, req, server_ids):
        for server_id in server_ids:
            self._release_floating_ips(req, server_id)
            self.os_helper.delete(req, server_id)
        return []

    def delete(self, req, id):
        return self._delete(req, [id])

    def delete_all(self, req):
        ids = [s["id"] for s in self.os_helper.index(req)]
        return self._delete(req, ids)
